# Copyrigh 2016 Cisco Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import collections
import copy

import mock

from neutron.tests import base as base_test

from networking_cisco.apps.saf.server.services.firewall.native import (
    fabric_setup_base as FP)
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    base)
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    native)
import networking_cisco.apps.saf.server.services.firewall.native.fw_constants \
    as fw_const


TIME = 'Thu Aug 11 19:11:30 2000'
TENANT_NAME = 'TenantA'
TENANT_ID = '0000-1111-2222-5555'
FW_ID = '0000-aaaa-bbbb-ccce'
FW_NAME = 'FwA'
POLCY_ID = '0000-aaaa-bbbb-cccc'
FW_TYPE = 'TE'
ROUTER_ID = '0000-aaaa-bbbb-5555'
RULE_ID = '0000-aaaa-bbbb-cccd'
NET_ID = '0000-aaa1-bbbb-cccd'
OUT_NET_ID = '0000-aab1-bbbb-cccd'
RTR_ID = '0000-aaa2-bbbb-cccd'
SUBNET_ID = '0000-aaa3-bbbb-cccd'
OUT_SUBNET_ID = '0000-aaa4-bbbb-cccd'
PROTOCOL = 'tcp'
MGMT_IP = '3.3.3.3'
SRC_IP = '1.1.1.1'
DST_IP = '2.2.2.2'
NEW_CIDR = '10.11.12.0/24'
SRC_PORT = 34
DST_PORT = 43
RULE_NAME = 'RuleA'
IN_SUBNET = '100.100.11.0'
IN_MASK = '255.255.255.0'
IN_SUBNET_AND_MASK = '100.100.11.0/24'
IN_START = '100.100.11.3'
IN_SEC_GW = '100.100.11.254'
IN_GW = '100.100.11.2'
IN_FABRIC_GW = '100.100.11.1'
IN_END = '100.100.11.254'
SEGMENTATION_ID = 87999
OUT_SEGMENTATION_ID = 88000
VLAN_ID = 770
OUT_VLAN_ID = 771
OUT_SUBNET = '200.200.11.0'
OUT_MASK = '255.255.255.0'
OUT_SUBNET_AND_MASK = '200.200.11.0/24'
RTR_SUBNET_AND_MASK = '9.9.9.0/24'
OUT_START = '200.200.11.3'
OUT_SEC_GW = '200.200.11.254'
OUT_GW = '200.200.11.2'
OUT_FABRIC_GW = '200.200.11.1'
OUT_END = '200.200.11.254'
INTF_IN = 'e1/1'
INTF_OUT = 'e1/2'
HOST_NAME = 'Host1'
VM_MAC = '00:00:00:11:22:33'
VM_IP = '2.3.5.4'
PORT_ID = '0000-aab3-bbbb-cccd'
STATIC_SUB = '2.3.5.0'
NEW_STATIC_SUB = '10.11.12.0'

try:
    OrderedDict = collections.OrderedDict
except AttributeError:
    import ordereddict
    OrderedDict = ordereddict.OrderedDict


class FakeClass(object):
    """Fake class"""
    @classmethod
    def imitate(cls, *others):
        for other in others:
            for name in other.__dict__:
                try:
                    setattr(cls, name, mock.Mock())
                except (TypeError, AttributeError):
                    pass
        return cls

    @classmethod
    def set_return(cls, class_name, fn_name, return_val):
        getattr(cls, fn_name).return_value = return_val


class NativeFirewallTest(base_test.BaseTestCase):
    """A test suite to exercise the Native Firewall Driver.  """

    def setUp(self):
        '''Setup for the test scripts '''
        super(NativeFirewallTest, self).setUp()
        self._init_values()
        self.cfg_dict = self._fill_cfg()

        native.NativeFirewall.__bases__ = (FakeClass.imitate(base.BaseDriver,
                                                             FP.FabricApi),)
        mock.patch('networking_cisco.apps.saf.server.'
                   'dfa_openstack_helper.DfaNeutronHelper').start()
        mock.patch('time.sleep').start()
        mock.patch('time.ctime', return_value=TIME).start()
        self.native_fw = native.NativeFirewall()
        self.native_fw.initialize(self.cfg_dict)
        self.native_fw.populate_dcnm_obj(mock.MagicMock())
        self.native_fw.populate_event_que(mock.MagicMock())

    def _init_values(self):
        self.tenant_name = TENANT_NAME
        self.tenant_id = TENANT_ID
        self.rule_id = RULE_ID
        self.in_subnet = IN_SUBNET
        self.in_subnet_and_mask = IN_SUBNET_AND_MASK
        self.in_gw = IN_GW
        self.in_start = IN_START
        self.in_sec_gw = IN_SEC_GW
        self.in_end = IN_END
        self.out_subnet = OUT_SUBNET
        self.out_subnet_and_mask = OUT_SUBNET_AND_MASK
        self.out_gw = OUT_GW
        self.out_sec_gw = OUT_SEC_GW
        self.out_start = OUT_START
        self.out_end = OUT_END
        self.in_fabric_gw = IN_FABRIC_GW
        self.out_fabric_gw = OUT_FABRIC_GW
        self.segmentation_id = SEGMENTATION_ID
        self.out_segmentation_id = OUT_SEGMENTATION_ID
        self.vlan_id = VLAN_ID
        self.out_vlan_id = OUT_VLAN_ID
        self.net_id = NET_ID
        self.out_net_id = OUT_NET_ID
        self.rtr_id = RTR_ID
        self.subnet_id = SUBNET_ID
        self.out_subnet_id = OUT_SUBNET_ID
        self.fw_data = {'tenant_name': self.tenant_name,
                        'router_id': self.rtr_id}

    def _fill_cfg(self):
        config = {'mgmt_ip_addr': MGMT_IP, 'user': 'user',
                  'pwd': 'user', 'interface_in': INTF_IN,
                  'interface_out': INTF_OUT}
        return config

    def test_native_fw_init(self):
        '''Wrapper for the init'''
        pass

    def _get_modified_fw_data(self):
        new_fw_data = copy.deepcopy(self.fw_data)
        new_fw_data.get('rules').get(self.rule_id)['action'] = 'deny'
        return new_fw_data

    def _get_port_data(self):
        port_data = {'binding:host_id': HOST_NAME, 'mac_address': VM_MAC,
                     'id': PORT_ID, 'fixed_ips': [{'ip_address': VM_IP}]}
        return port_data

    def _get_vm_vdp_data(self, status, net_id, segid, oper, event):
        vm_name = 'FW_SRVC_RTR_' + self.tenant_name + '_' + oper
        vm_vdp_data = {'status': status, 'vm_name': vm_name,
                       'network_id': net_id, 'gw_mac': None, 'host': HOST_NAME,
                       'segid': segid, 'mac': VM_MAC, 'vm_uuid': self.rtr_id,
                       'vm_ip': VM_IP, 'port_id': PORT_ID,
                       'fwd_mod': 'anycast_gateway'}
        payload = {'service': vm_vdp_data}
        data = (event, payload)
        return data

    def _test_send_in_router_port_msg(self, status):
        """Test for the 'in' service vNic create message. """
        port_data = self._get_port_data()
        if status == 'up':
            msg = 'service.vnic.create'
        else:
            msg = 'service.vnic.delete'
        vm_vdp_data_in = self._get_vm_vdp_data(status, self.net_id,
                                               self.segmentation_id,
                                               'in', msg)
        with mock.patch.object(self.native_fw.que_obj, 'put') as que_put,\
            mock.patch.object(self.native_fw.os_helper,
                              'get_router_port_subnet',
                              return_value=port_data):
            FakeClass.set_return(FP.FabricApi, 'get_in_subnet_id',
                                 self.subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_subnet_id',
                                 self.out_subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_in_net_id', self.net_id)
            parent = mock.MagicMock()
            parent.attach_mock(que_put, 'put')
            arg_dict = self.native_fw._create_arg_dict(self.tenant_id,
                                                       self.fw_data,
                                                       self.subnet_id,
                                                       self.out_subnet_id)
            self.native_fw.tenant_dict[self.tenant_id] = (
                {'host': None, 'router_id': self.rtr_id})
            self.native_fw.send_in_router_port_msg(self.tenant_id, arg_dict,
                                                   status)
        expected_calls = [mock.call.put((34, TIME, vm_vdp_data_in))]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_send_in_router_port_msg_up(self):
        self._test_send_in_router_port_msg('up')

    def test_send_in_router_port_msg_down(self):
        self._test_send_in_router_port_msg('down')

    def _test_send_in_router_port_msg_fail(self, status):
        """Test for the 'in' service vNic create message fail. """
        subnet_lst = set()
        subnet_lst.add(self.subnet_id)
        subnet_lst.add(self.out_subnet_id)
        with mock.patch.object(self.native_fw.que_obj, 'put') as que_put,\
            mock.patch.object(self.native_fw.os_helper,
                              'get_router_port_subnet',
                              return_value=None),\
            mock.patch.object(self.native_fw.os_helper,
                              'delete_intf_router',
                              return_value=True) as del_intf_rtr:
            FakeClass.set_return(FP.FabricApi, 'get_in_subnet_id',
                                 self.subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_subnet_id',
                                 self.out_subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_in_net_id', self.net_id)
            parent = mock.MagicMock()
            parent.attach_mock(que_put, 'put')
            if status == 'up':
                parent.attach_mock(del_intf_rtr, 'delete_intf_router')
            arg_dict = self.native_fw._create_arg_dict(self.tenant_id,
                                                       self.fw_data,
                                                       self.subnet_id,
                                                       self.out_subnet_id)
            self.native_fw.tenant_dict[self.tenant_id] = (
                {'host': None, 'router_id': self.rtr_id})
            self.native_fw.send_in_router_port_msg(self.tenant_id, arg_dict,
                                                   status)
        expected_calls = []
        if status == 'up':
            expected_calls.append(mock.call.delete_intf_router(
                self.tenant_name, self.tenant_id, self.rtr_id, subnet_lst))
        parent.assert_has_calls(expected_calls, any_order=False)
        self.assertEqual(que_put.called, False)

    def test_send_in_router_port_msg_fail_up(self):
        self._test_send_in_router_port_msg_fail('up')

    def test_send_in_router_port_msg_fail_down(self):
        self._test_send_in_router_port_msg_fail('down')

    def _test_send_out_router_port_msg(self, status):
        """Test for the 'out' service vNic create message. """
        port_data = self._get_port_data()
        if status == 'up':
            msg = 'service.vnic.create'
        else:
            msg = 'service.vnic.delete'
        vm_vdp_data_out = self._get_vm_vdp_data(status, self.out_net_id,
                                                self.out_segmentation_id,
                                                'out', msg)
        with mock.patch.object(self.native_fw.que_obj, 'put') as que_put,\
            mock.patch.object(self.native_fw.os_helper,
                              'get_router_port_subnet',
                              return_value=port_data):
            FakeClass.set_return(FP.FabricApi, 'get_in_subnet_id',
                                 self.subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_subnet_id',
                                 self.out_subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_net_id',
                                 self.out_net_id)
            parent = mock.MagicMock()
            parent.attach_mock(que_put, 'put')
            arg_dict = self.native_fw._create_arg_dict(self.tenant_id,
                                                       self.fw_data,
                                                       self.subnet_id,
                                                       self.out_subnet_id)
            self.native_fw.tenant_dict[self.tenant_id] = (
                {'host': None, 'router_id': self.rtr_id})
            self.native_fw.send_out_router_port_msg(self.tenant_id, arg_dict,
                                                    status)
        expected_calls = [mock.call.put((34, TIME, vm_vdp_data_out))]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_send_out_router_port_msg_up(self):
        self._test_send_out_router_port_msg('up')

    def test_send_out_router_port_msg_down(self):
        self._test_send_out_router_port_msg('down')

    def _test_send_out_router_port_msg_fail(self, status):
        """Test for the 'out' service vNic create message fail. """
        subnet_lst = set()
        subnet_lst.add(self.subnet_id)
        subnet_lst.add(self.out_subnet_id)
        with mock.patch.object(self.native_fw.que_obj, 'put') as que_put,\
            mock.patch.object(self.native_fw.os_helper,
                              'get_router_port_subnet',
                              return_value=None),\
            mock.patch.object(self.native_fw.os_helper,
                              'delete_intf_router',
                              return_value=True) as del_intf_rtr:
            FakeClass.set_return(FP.FabricApi, 'get_in_subnet_id',
                                 self.subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_subnet_id',
                                 self.out_subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_net_id', self.net_id)
            parent = mock.MagicMock()
            parent.attach_mock(que_put, 'put')
            if status == 'up':
                parent.attach_mock(del_intf_rtr, 'delete_intf_router')
            arg_dict = self.native_fw._create_arg_dict(self.tenant_id,
                                                       self.fw_data,
                                                       self.subnet_id,
                                                       self.out_subnet_id)
            self.native_fw.tenant_dict[self.tenant_id] = (
                {'host': None, 'router_id': self.rtr_id})
            self.native_fw.send_out_router_port_msg(self.tenant_id, arg_dict,
                                                    status)
        expected_calls = []
        if status == 'up':
            expected_calls.append(mock.call.delete_intf_router(
                self.tenant_name, self.tenant_id, self.rtr_id, subnet_lst))
        parent.assert_has_calls(expected_calls, any_order=False)
        self.assertEqual(que_put.called, False)

    def test_send_out_router_port_msg_fail_up(self):
        self._test_send_out_router_port_msg_fail('up')

    def test_send_out_router_port_msg_fail_down(self):
        self._test_send_out_router_port_msg_fail('down')

    def test_program_next_hop(self):
        """Test for programming the next hop for out network. """
        with mock.patch.object(self.native_fw.os_helper,
                               'program_rtr_all_nwk_next_hop',
                               return_value=True) as prog_rtr_all_nwk:
            FakeClass.set_return(FP.FabricApi, 'get_in_ip_addr',
                                 {'subnet': self.in_subnet,
                                  'start': self.in_start,
                                  'sec_gateway': self.in_sec_gw,
                                  'gateway': self.in_fabric_gw,
                                  'end': self.in_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            parent = mock.MagicMock()
            parent.attach_mock(prog_rtr_all_nwk,
                               'program_rtr_all_nwk_next_hop')
            arg_dict = self.native_fw._create_arg_dict(self.tenant_id,
                                                       self.fw_data,
                                                       self.subnet_id,
                                                       self.out_subnet_id)
            self.native_fw.program_next_hop(self.tenant_id, arg_dict)
        expected_calls = [
            mock.call.program_rtr_all_nwk_next_hop(
                self.tenant_id, self.rtr_id, self.in_fabric_gw,
                [self.in_subnet, self.out_subnet])]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_program_next_hop_fail(self):
        """Test for programming the next hop for out network, failure case. """
        subnet_lst = set()
        subnet_lst.add(self.subnet_id)
        subnet_lst.add(self.out_subnet_id)
        with mock.patch.object(self.native_fw.os_helper,
                               'program_rtr_all_nwk_next_hop',
                               return_value=False) as prog_rtr_all_nwk,\
            mock.patch.object(self.native_fw.os_helper,
                              'delete_intf_router',
                              return_value=True) as del_intf_rtr:
            FakeClass.set_return(FP.FabricApi, 'get_in_ip_addr',
                                 {'subnet': self.in_subnet,
                                  'start': self.in_start,
                                  'sec_gateway': self.in_sec_gw,
                                  'gateway': self.in_fabric_gw,
                                  'end': self.in_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_in_subnet_id',
                                 self.subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_subnet_id',
                                 self.out_subnet_id)
            parent = mock.MagicMock()
            parent.attach_mock(prog_rtr_all_nwk,
                               'program_rtr_all_nwk_next_hop')
            parent.attach_mock(del_intf_rtr, 'delete_intf_router')
            self.native_fw.tenant_dict[self.tenant_id] = (
                {'host': None, 'router_id': self.rtr_id})
            arg_dict = self.native_fw._create_arg_dict(self.tenant_id,
                                                       self.fw_data,
                                                       self.subnet_id,
                                                       self.out_subnet_id)
            self.native_fw.program_next_hop(self.tenant_id, arg_dict)
        expected_calls = [
            mock.call.program_rtr_all_nwk_next_hop(
                self.tenant_id, self.rtr_id, self.in_fabric_gw,
                [self.in_subnet, self.out_subnet]),
            mock.call.delete_intf_router(self.tenant_name,
                                         self.tenant_id,
                                         self.rtr_id,
                                         subnet_lst)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_program_default_gw(self):
        """Test for programming the default GW. """
        with mock.patch.object(self.native_fw.os_helper,
                               'program_rtr_default_gw',
                               return_value=True) as prog_def_gw:
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            parent = mock.MagicMock()
            parent.attach_mock(prog_def_gw, 'program_rtr_default_gw')
            arg_dict = self.native_fw._create_arg_dict(self.tenant_id,
                                                       self.fw_data,
                                                       self.subnet_id,
                                                       self.out_subnet_id)
            self.native_fw.program_default_gw(self.tenant_id, arg_dict)
        expected_calls = [
            mock.call.program_rtr_default_gw(
                self.tenant_id, self.rtr_id, self.out_fabric_gw)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_program_default_gw_fail(self):
        """Test for programming the default gw, failure case. """
        subnet_lst = set()
        subnet_lst.add(self.subnet_id)
        subnet_lst.add(self.out_subnet_id)
        with mock.patch.object(self.native_fw.os_helper,
                               'program_rtr_default_gw',
                               return_value=False) as prog_def_gw,\
            mock.patch.object(self.native_fw.os_helper,
                              'delete_intf_router',
                              return_value=True) as del_intf_rtr:
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_in_subnet_id',
                                 self.subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_subnet_id',
                                 self.out_subnet_id)
            parent = mock.MagicMock()
            parent.attach_mock(prog_def_gw, 'program_rtr_default_gw')
            parent.attach_mock(del_intf_rtr, 'delete_intf_router')
            self.native_fw.tenant_dict[self.tenant_id] = (
                {'host': None, 'router_id': self.rtr_id})
            arg_dict = self.native_fw._create_arg_dict(self.tenant_id,
                                                       self.fw_data,
                                                       self.subnet_id,
                                                       self.out_subnet_id)
            self.native_fw.program_default_gw(self.tenant_id, arg_dict)
        expected_calls = [
            mock.call.program_rtr_default_gw(
                self.tenant_id, self.rtr_id, self.out_fabric_gw),
            mock.call.delete_intf_router(self.tenant_name,
                                         self.tenant_id,
                                         self.rtr_id,
                                         subnet_lst)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_update_dcnm_part_static_route(self):
        """Test for updating the DCNM partition's static route. """
        with mock.patch.object(self.native_fw.os_helper,
                               'get_subnet_nwk_excl',
                               return_value=[STATIC_SUB]),\
            mock.patch.object(self.native_fw.dcnm_obj,
                              'update_partition_static_route',
                              return_value=True) as upd_part:
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_srvc_node_ip_addr',
                                 self.out_gw)
            parent = mock.MagicMock()
            parent.attach_mock(upd_part, 'update_partition_static_route')
            arg_dict = self.native_fw._create_arg_dict(self.tenant_id,
                                                       self.fw_data,
                                                       self.subnet_id,
                                                       self.out_subnet_id)
            self.native_fw.update_dcnm_partition_static_route(
                self.tenant_id, arg_dict)
        expected_calls = [
            mock.call.update_partition_static_route(
                self.tenant_name, fw_const.SERV_PART_NAME, [STATIC_SUB],
                vrf_prof=(
                    self.native_fw.cfg.firewall.fw_service_part_vrf_profile),
                service_node_ip=self.out_gw)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_update_dcnm_part_static_route_fail(self):
        """Test for updating the DCNM partition's static route. """
        subnet_lst = set()
        subnet_lst.add(self.subnet_id)
        subnet_lst.add(self.out_subnet_id)
        with mock.patch.object(self.native_fw.os_helper,
                               'get_subnet_nwk_excl',
                               return_value=[STATIC_SUB]),\
            mock.patch.object(self.native_fw.dcnm_obj,
                              'update_partition_static_route',
                              return_value=False) as upd_part,\
            mock.patch.object(self.native_fw.os_helper,
                              'delete_intf_router',
                              return_value=True) as del_intf_rtr:
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_in_subnet_id',
                                 self.subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_subnet_id',
                                 self.out_subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_srvc_node_ip_addr',
                                 self.out_gw)
            parent = mock.MagicMock()
            parent.attach_mock(upd_part, 'update_partition_static_route')
            parent.attach_mock(del_intf_rtr, 'delete_intf_router')
            self.native_fw.tenant_dict[self.tenant_id] = (
                {'host': None, 'router_id': self.rtr_id})
            arg_dict = self.native_fw._create_arg_dict(self.tenant_id,
                                                       self.fw_data,
                                                       self.subnet_id,
                                                       self.out_subnet_id)
            self.native_fw.update_dcnm_partition_static_route(
                self.tenant_id, arg_dict)
        expected_calls = [
            mock.call.update_partition_static_route(
                self.tenant_name, fw_const.SERV_PART_NAME, [STATIC_SUB],
                vrf_prof=(
                    self.native_fw.cfg.firewall.fw_service_part_vrf_profile),
                service_node_ip=self.out_gw),
            mock.call.delete_intf_router(self.tenant_name,
                                         self.tenant_id,
                                         self.rtr_id,
                                         subnet_lst)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def _test_create_fw(self):
        port_data = self._get_port_data()
        subnet_lst = set()
        subnet_lst.add(self.subnet_id)
        subnet_lst.add(self.out_subnet_id)
        vm_vdp_data_in = self._get_vm_vdp_data('up', self.net_id,
                                               self.segmentation_id,
                                               'in', 'service.vnic.create')
        vm_vdp_data_out = self._get_vm_vdp_data('up', self.out_net_id,
                                                self.out_segmentation_id,
                                                'out', 'service.vnic.create')
        with mock.patch.object(self.native_fw.os_helper, 'add_intf_router',
                               return_value=True) as add_intf_rtr,\
                mock.patch.object(self.native_fw.os_helper,
                                  'get_subnet_nwk_excl',
                                  return_value=[STATIC_SUB]),\
                mock.patch.object(self.native_fw.dcnm_obj,
                                  'update_partition_static_route',
                                  return_value=True) as upd_part,\
                mock.patch.object(self.native_fw.os_helper,
                                  'program_rtr_default_gw',
                                  return_value=True) as prog_def_gw,\
                mock.patch.object(self.native_fw.os_helper,
                                  'program_rtr_all_nwk_next_hop',
                                  return_value=True) as prog_rtr_all_nwk,\
                mock.patch.object(self.native_fw.os_helper,
                                  'get_router_port_subnet',
                                  return_value=port_data),\
                mock.patch.object(self.native_fw.que_obj, 'put') as que_put:
            FakeClass.set_return(FP.FabricApi, 'get_in_ip_addr',
                                 {'subnet': self.in_subnet,
                                  'start': self.in_start,
                                  'sec_gateway': self.in_sec_gw,
                                  'gateway': self.in_fabric_gw,
                                  'end': self.in_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_in_srvc_node_ip_addr',
                                 self.in_gw)
            FakeClass.set_return(FP.FabricApi, 'get_out_srvc_node_ip_addr',
                                 self.out_gw)
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_in_net_id', self.net_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_net_id',
                                 self.out_net_id)
            FakeClass.set_return(FP.FabricApi, 'get_in_subnet_id',
                                 self.subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_subnet_id',
                                 self.out_subnet_id)
            parent = mock.MagicMock()
            parent.attach_mock(add_intf_rtr, 'add_intf_router')
            parent.attach_mock(upd_part, 'update_partition_static_route')
            parent.attach_mock(prog_def_gw, 'program_rtr_default_gw')
            parent.attach_mock(prog_rtr_all_nwk,
                               'program_rtr_all_nwk_next_hop')
            parent.attach_mock(que_put, 'put')
            self.native_fw.create_fw(self.tenant_id, self.fw_data)
        expected_calls = [
            mock.call.add_intf_router(self.rtr_id, self.tenant_id, subnet_lst),
            mock.call.update_partition_static_route(
                self.tenant_name, fw_const.SERV_PART_NAME, [STATIC_SUB],
                vrf_prof=(
                    self.native_fw.cfg.firewall.fw_service_part_vrf_profile),
                service_node_ip=self.out_gw),
            mock.call.program_rtr_default_gw(self.tenant_id, self.rtr_id,
                                             self.out_fabric_gw),
            mock.call.program_rtr_all_nwk_next_hop(
                self.tenant_id, self.rtr_id, self.in_fabric_gw,
                [self.in_subnet, self.out_subnet]),
            mock.call.put((34, TIME, vm_vdp_data_in)),
            mock.call.put((34, TIME, vm_vdp_data_out))]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_create_fw(self):
        """Create FW Test """
        self._test_create_fw()

    def _test_delete_fw(self):
        port_data = self._get_port_data()
        subnet_lst = set()
        subnet_lst.add(self.subnet_id)
        subnet_lst.add(self.out_subnet_id)
        vm_vdp_data_in = self._get_vm_vdp_data('down', self.net_id,
                                               self.segmentation_id,
                                               'in', 'service.vnic.delete')
        vm_vdp_data_out = self._get_vm_vdp_data('down', self.out_net_id,
                                                self.out_segmentation_id,
                                                'out', 'service.vnic.delete')
        with mock.patch.object(self.native_fw.os_helper,
                               'delete_intf_router',
                               return_value=True) as del_intf_rtr,\
                mock.patch.object(self.native_fw.os_helper,
                                  'get_router_port_subnet',
                                  return_value=port_data),\
                mock.patch.object(self.native_fw.que_obj, 'put') as que_put:

            FakeClass.set_return(
                FP.FabricApi, 'get_in_ip_addr',
                {'subnet': self.in_subnet, 'start': self.in_start,
                 'sec_gateway': self.in_sec_gw, 'gateway': self.in_fabric_gw,
                 'end': self.in_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_in_seg_vlan',
                                 [self.segmentation_id, self.vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_out_seg_vlan',
                                 [self.out_segmentation_id, self.out_vlan_id])
            FakeClass.set_return(FP.FabricApi, 'get_in_net_id', self.net_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_net_id',
                                 self.out_net_id)
            FakeClass.set_return(FP.FabricApi, 'get_in_subnet_id',
                                 self.subnet_id)
            FakeClass.set_return(FP.FabricApi, 'get_out_subnet_id',
                                 self.out_subnet_id)
            parent = mock.MagicMock()
            parent.attach_mock(que_put, 'put')
            parent.attach_mock(del_intf_rtr, 'delete_intf_router')
            self.native_fw.delete_fw(self.tenant_id, self.fw_data)
        expected_calls = [
            mock.call.put((34, TIME, vm_vdp_data_in)),
            mock.call.put((34, TIME, vm_vdp_data_out)),
            mock.call.delete_intf_router(self.tenant_name, self.tenant_id,
                                         self.rtr_id, subnet_lst)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_delete_fw(self):
        """Delete FW Test """
        self._test_delete_fw()

    def _test_nwk_create_notif(self):
        subnet_lst = set()
        subnet_lst.add(self.subnet_id)
        subnet_lst.add(self.out_subnet_id)
        with mock.patch.object(self.native_fw.os_helper, 'get_subnet_nwk_excl',
                               return_value=[NEW_STATIC_SUB]),\
            mock.patch.object(self.native_fw.os_helper, 'get_rtr_by_name',
                              return_value=[{'id': self.rtr_id}]),\
            mock.patch.object(self.native_fw.dcnm_obj,
                              'update_partition_static_route',
                              return_value=True) as upd_part,\
            mock.patch.object(self.native_fw.os_helper,
                              'program_rtr_nwk_next_hop',
                              return_value=True) as prog_rtr_nwk:
            FakeClass.set_return(FP.FabricApi, 'get_in_ip_addr',
                                 {'subnet': self.in_subnet,
                                  'start': self.in_start,
                                  'sec_gateway': self.in_sec_gw,
                                  'gateway': self.in_fabric_gw,
                                  'end': self.in_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_srvc_node_ip_addr',
                                 self.out_gw)
            parent = mock.MagicMock()
            parent.attach_mock(upd_part, 'update_partition_static_route')
            parent.attach_mock(prog_rtr_nwk, 'program_rtr_nwk_next_hop')
            self.native_fw.network_create_notif(self.tenant_id,
                                                self.tenant_name, NEW_CIDR)
        expected_calls = [
            mock.call.update_partition_static_route(
                self.tenant_name, fw_const.SERV_PART_NAME, [NEW_STATIC_SUB],
                vrf_prof=(
                    self.native_fw.cfg.firewall.fw_service_part_vrf_profile),
                service_node_ip=self.out_gw),
            mock.call.program_rtr_nwk_next_hop(self.rtr_id, self.in_fabric_gw,
                                               NEW_CIDR)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_nwk_create_notif(self):
        """Nwk Create Notif """
        self._test_nwk_create_notif()

    def _test_nwk_delete_notif(self):
        excl_lst = []
        excl_lst.append(self.in_subnet)
        excl_lst.append(self.out_subnet)
        with mock.patch.object(self.native_fw.os_helper, 'get_subnet_nwk_excl',
                               return_value=[NEW_STATIC_SUB]),\
            mock.patch.object(self.native_fw.os_helper, 'get_rtr_by_name',
                              return_value=[{'id': self.rtr_id}]),\
            mock.patch.object(self.native_fw.dcnm_obj,
                              'update_partition_static_route',
                              return_value=True) as upd_part,\
            mock.patch.object(self.native_fw.os_helper,
                              'remove_rtr_nwk_next_hop',
                              return_value=True) as prog_rtr_nwk:
            FakeClass.set_return(FP.FabricApi, 'get_in_ip_addr',
                                 {'subnet': self.in_subnet,
                                  'start': self.in_start,
                                  'sec_gateway': self.in_sec_gw,
                                  'gateway': self.in_fabric_gw,
                                  'end': self.in_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_ip_addr',
                                 {'subnet': self.out_subnet,
                                  'start': self.out_start,
                                  'sec_gateway': self.out_sec_gw,
                                  'gateway': self.out_fabric_gw,
                                  'end': self.out_end})
            FakeClass.set_return(FP.FabricApi, 'get_out_srvc_node_ip_addr',
                                 self.out_gw)
            parent = mock.MagicMock()
            parent.attach_mock(upd_part, 'update_partition_static_route')
            parent.attach_mock(prog_rtr_nwk, 'remove_rtr_nwk_next_hop')
            self.native_fw.network_delete_notif(self.tenant_id,
                                                self.tenant_name, NEW_CIDR)
        expected_calls = [
            mock.call.update_partition_static_route(
                self.tenant_name, fw_const.SERV_PART_NAME, [NEW_STATIC_SUB],
                vrf_prof=(
                    self.native_fw.cfg.firewall.fw_service_part_vrf_profile),
                service_node_ip=self.out_gw),
            mock.call.remove_rtr_nwk_next_hop(self.rtr_id, self.in_fabric_gw,
                                              [NEW_STATIC_SUB], excl_lst)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_nwk_delete_notif(self):
        """Create Notif """
        self._test_nwk_delete_notif()



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import mock
import unittest

from neutron.api.rpc.agentnotifiers import l3_rpc_agent_api
from neutron.common import test_lib
from neutron.tests import fake_notifier
from neutron.tests.unit.db import test_agentschedulers_db
from neutron.tests.unit.extensions import test_l3
from neutron.tests.unit.scheduler import test_l3_agent_scheduler
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_utils import importutils
from oslo_utils import uuidutils
from sqlalchemy import exc as inner_db_exc
from webob import exc

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import agent_exceptions
from networking_cisco.backwards_compatibility import l3_const
from networking_cisco.plugins.cisco.common import cisco_constants as c_const
from networking_cisco.plugins.cisco.db.l3 import ha_db
from networking_cisco.plugins.cisco.db.scheduler import (
    l3_routertype_aware_schedulers_db as router_sch_db)
from networking_cisco.plugins.cisco.extensions import ciscohostingdevicemanager
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.plugins.cisco.extensions import routerhostingdevice
from networking_cisco.plugins.cisco.extensions import routertype
from networking_cisco.plugins.cisco.extensions import routertypeawarescheduler
from networking_cisco.plugins.cisco.l3.rpc import l3_router_rpc_cfg_agent_api
from networking_cisco.plugins.cisco.l3.schedulers import (
    l3_router_hosting_device_scheduler as scheduler)
from networking_cisco.tests.unit.cisco.device_manager import (
    device_manager_test_support)
from networking_cisco.tests.unit.cisco.device_manager import (
    test_db_device_manager)
from networking_cisco.tests.unit.cisco.l3 import (
    test_l3_router_appliance_plugin)
from networking_cisco.tests.unit.cisco.l3 import l3_router_test_support
from networking_cisco.tests.unit.cisco.l3 import test_db_routertype


CORE_PLUGIN_KLASS = device_manager_test_support.CORE_PLUGIN_KLASS
L3_PLUGIN_KLASS = (
    'networking_cisco.tests.unit.cisco.l3.test_l3_routertype_aware_schedulers.'
    'TestSchedulingCapableL3RouterServicePlugin')
HA_L3_PLUGIN_KLASS = ('networking_cisco.tests.unit.cisco.l3.'
                      'test_l3_routertype_aware_schedulers.'
                      'TestSchedulingHACapableL3RouterServicePlugin')

_uuid = uuidutils.generate_uuid
HOSTING_DEVICE_ATTR = routerhostingdevice.HOSTING_DEVICE_ATTR
HARDWARE_CATEGORY = ciscohostingdevicemanager.HARDWARE_CATEGORY
AGENT_TYPE_L3_CFG = c_const.AGENT_TYPE_L3_CFG
NEUTRON_VERSION = bc.NEUTRON_VERSION
NEUTRON_NEWTON_VERSION = bc.NEUTRON_NEWTON_VERSION


class TestSchedulingL3RouterApplianceExtensionManager(
        test_db_routertype.L3TestRoutertypeExtensionManager):

    def get_resources(self):
        # first, add auto_schedule and share_hosting_device attributes to
        # router resource
        l3_const.RESOURCE_ATTRIBUTE_MAP['routers'].update(
            routertypeawarescheduler.EXTENDED_ATTRIBUTES_2_0['routers'])
        # most of the resources are added in our super class
        res = super(TestSchedulingL3RouterApplianceExtensionManager,
                    self).get_resources()
        # add the router to hosting device scheduler resources
        ext_mgr = routertypeawarescheduler.Routertypeawarescheduler()
        for item in ext_mgr.get_resources():
            res.append(item)
        return res


# A scheduler-enabled routertype capable L3 routing service plugin class
class TestSchedulingCapableL3RouterServicePlugin(
    l3_router_test_support.TestL3RouterServicePlugin,
        router_sch_db.L3RouterTypeAwareSchedulerDbMixin):

    supported_extension_aliases = (
        l3_router_test_support.TestL3RouterServicePlugin.
        supported_extension_aliases +
        [routertypeawarescheduler.ROUTERTYPE_AWARE_SCHEDULER_ALIAS,
         bc.constants.L3_AGENT_SCHEDULER_EXT_ALIAS])

    def __init__(self):
        self.agent_notifiers.update(
            {bc.constants.AGENT_TYPE_L3: l3_rpc_agent_api.L3AgentNotifyAPI(),
             c_const.AGENT_TYPE_L3_CFG:
             l3_router_rpc_cfg_agent_api.L3RouterCfgAgentNotifyAPI(self)})
        self.router_scheduler = importutils.import_object(
            cfg.CONF.routing.router_type_aware_scheduler_driver)
        self.l3agent_scheduler = importutils.import_object(
            cfg.CONF.router_scheduler_driver)

    def cleanup_after_test(self):
        """Reset all class variables to their default values.
        This is needed to avoid tests to pollute subsequent tests.
        """
        TestSchedulingCapableL3RouterServicePlugin._router_schedulers = {}
        TestSchedulingCapableL3RouterServicePlugin._router_drivers = {}
        (TestSchedulingCapableL3RouterServicePlugin.
         _namespace_router_type_id) = None
        TestSchedulingCapableL3RouterServicePlugin._backlogged_routers = set()
        (TestSchedulingCapableL3RouterServicePlugin.
         _refresh_router_backlog) = True


class L3RoutertypeAwareL3AgentSchedulerTestCase(
    test_l3_agent_scheduler.L3SchedulerTestCaseMixin,
    test_db_routertype.RoutertypeTestCaseMixin,
    test_db_device_manager.DeviceManagerTestCaseMixin,
    l3_router_test_support.L3RouterTestSupportMixin,
        device_manager_test_support.DeviceManagerTestSupportMixin):

    resource_prefix_map = (test_db_device_manager.TestDeviceManagerDBPlugin
                           .resource_prefix_map)

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None):
        if not core_plugin:
            core_plugin = CORE_PLUGIN_KLASS
        if l3_plugin is None:
            l3_plugin = L3_PLUGIN_KLASS
        service_plugins = {'l3_plugin_name': l3_plugin}

        cfg.CONF.set_override('api_extensions_path',
                              l3_router_test_support.extensions_path)
        if ext_mgr is None:
            ext_mgr = TestSchedulingL3RouterApplianceExtensionManager()

        # call grandparent's setUp() to avoid that wrong plugin and
        # extensions are used.
        super(test_l3_agent_scheduler.L3SchedulerTestCaseMixin, self).setUp(
            plugin=core_plugin, service_plugins=service_plugins,
            ext_mgr=ext_mgr)

        self._define_keystone_authtoken()
        cfg.CONF.set_override('default_router_type',
                              c_const.NAMESPACE_ROUTER_TYPE, group='routing')

        self.adminContext = bc.context.get_admin_context()
        self.plugin = bc.get_plugin()
        self.l3_plugin = bc.get_plugin(bc.constants.L3)
        # work-around to make some tests in super class, which assumes core
        # plugin does the l3 routing, run correctly
        self.plugin.router_scheduler = (
            self.l3_plugin.l3agent_scheduler)
        self._register_l3_agents()
        self._mock_l3_admin_tenant()
        templates = self._test_create_hosting_device_templates()
        self._test_create_routertypes(templates.values())
        # mock the periodic router backlog processing in the tests
        self._mock_backlog_processing(self.l3_plugin)

    def tearDown(self):
        self._test_remove_routertypes()
        self._test_remove_hosting_device_templates()
        self.l3_plugin.cleanup_after_test()
        self.plugin.cleanup_after_test()
        super(L3RoutertypeAwareL3AgentSchedulerTestCase, self).tearDown()

    def _test_add_router_to_l3_agent(self,
                                     distributed=False,
                                     already_scheduled=False,
                                     external_gw=None):
        # Parent test class run tests that use this function with L3 agent
        # notifier set to None so we do the same.
        self.l3_plugin.agent_notifiers[bc.constants.AGENT_TYPE_L3] = None
        super(L3RoutertypeAwareL3AgentSchedulerTestCase,
              self)._test_add_router_to_l3_agent()

    def test_add_router_to_l3_agent_dvr_to_snat(self):
        # Parent test class run tests that use this function with L3 agent
        # notifier set to None so we do the same.
        self.l3_plugin.agent_notifiers[bc.constants.AGENT_TYPE_L3] = None
        super(L3RoutertypeAwareL3AgentSchedulerTestCase,
              self).test_add_router_to_l3_agent_dvr_to_snat()

    def test_get_unscheduled_routers_only_returns_namespace_routers(self):
        self._create_mgmt_nw_for_tests(self.fmt)
        arg_list = (routertype.TYPE_ATTR, )
        kwargs = {routertype.TYPE_ATTR: test_db_routertype.HW_ROUTERTYPE_NAME}
        # router 1
        self._make_router(self.fmt, _uuid(), 'router1', arg_list=arg_list,
                          **kwargs)['router']
        # namespace-based routers
        with self.router(name='router2') as router2, self.router(
                name='router3') as router3:
            r2 = router2['router']
            r3 = router3['router']
            # router 4
            self._make_router(self.fmt, _uuid(), 'router2', arg_list=arg_list,
                              **kwargs)['router']
            if NEUTRON_VERSION.version[0] <= NEUTRON_NEWTON_VERSION.version[0]:
                routers = (
                    self.l3_plugin.router_scheduler._get_unscheduled_routers(
                        self.adminContext, self.l3_plugin))
            else:
                routers = (
                    self.l3_plugin.router_scheduler._get_unscheduled_routers(
                        self.l3_plugin, self.adminContext))
            r_ids = set(r['id'] for r in routers)
            self.assertEqual(2, len(r_ids))
            for r in [r2, r3]:
                self.assertIn(r['id'], r_ids)
        self._remove_mgmt_nw_for_tests()

    def test_only_namespace_routers_scheduled_by_l3agent_scheduler(self):
        self._create_mgmt_nw_for_tests(self.fmt)
        arg_list = (routertype.TYPE_ATTR, )
        kwargs = {routertype.TYPE_ATTR: test_db_routertype.HW_ROUTERTYPE_NAME}
        r1 = self._make_router(self.fmt, _uuid(), 'router1',
                               arg_list=arg_list, **kwargs)['router']
        # namespace-based routers
        with self.router(name='router2') as router2,\
                self.router(name='router3') as router3,\
                mock.patch.object(self.l3_plugin.l3agent_scheduler,
                                  'schedule') as scheduler_mock,\
                mock.patch('neutron.scheduler.l3_agent_scheduler.L3Scheduler.'
                           '_get_routers_can_schedule') as auto_scheduler_mock:
            r2 = router2['router']
            r3 = router3['router']
            r4 = self._make_router(self.fmt, _uuid(), 'router4',
                                   arg_list=arg_list, **kwargs)['router']
            r_list = [r1, r2, r3, r4]
            # first test schedule function
            self.l3_plugin.schedule_routers(self.adminContext, r_list)
            scheduler_mock.assert_has_calls(
                [mock.call(mock.ANY, self.adminContext, r2['id'], None),
                 mock.call(mock.ANY, self.adminContext, r3['id'], None)])
            r_ids = [r['id'] for r in r_list]
            # then test auto_schedule function
            bc.auto_schedule_routers(self, 'host_1', r_ids)
            r_ids_scheduled = set([r['id'] for r in
                                   auto_scheduler_mock.call_args[0][2]])
            self.assertEqual(2, len(r_ids_scheduled))
            for r in [r2, r3]:
                self.assertIn(r['id'], r_ids_scheduled)
        self._remove_mgmt_nw_for_tests()

    def test_rpc_sync_routers_gets_only_namespace_routers(self):
        self._create_mgmt_nw_for_tests(self.fmt)
        arg_list = (routertype.TYPE_ATTR, )
        kwargs = {routertype.TYPE_ATTR: test_db_routertype.HW_ROUTERTYPE_NAME}
        # router 1
        self._make_router(self.fmt, _uuid(), 'router1', arg_list=arg_list,
                          **kwargs)['router']
        # namespace-based routers
        with self.router(name='router2') as router2, self.router(
                name='router3') as router3:
            r2 = router2['router']
            r3 = router3['router']
            # router 4
            self._make_router(self.fmt, _uuid(), 'router2', arg_list=arg_list,
                              **kwargs)['router']
            routers = self.l3_plugin.get_sync_data(self.adminContext)
            r_ids = set(r['id'] for r in routers)
            self.assertEqual(2, len(r_ids))
            for r in [r2, r3]:
                self.assertIn(r['id'], r_ids)
        self._remove_mgmt_nw_for_tests()

    def test_check_ports_exist_on_l3agent_with_dhcp_enabled_subnets(self):
        # overload to disable this test that fails as as it pertains to DVR
        # which we don't support
        pass


class L3RoutertypeAwareChanceL3AgentSchedulerTestCase(
    test_l3_agent_scheduler.L3AgentChanceSchedulerTestCase,
        L3RoutertypeAwareL3AgentSchedulerTestCase):

    def setUp(self):
        cfg.CONF.set_override('router_scheduler_driver',
                              'neutron.scheduler.l3_agent_scheduler.'
                              'ChanceScheduler')
        # call grandparent's setUp() to avoid that wrong scheduler is used
        super(test_l3_agent_scheduler.L3AgentChanceSchedulerTestCase,
              self).setUp()
        # Some UTs in parent class expects self.plugin to refer to l3 plugin
        self.plugin = self.l3_plugin

    def test_scheduler_auto_schedule_when_agent_added(self):
        # in our test setup the auto_schedule_routers function is provided by
        # the separate l3 service plugin, not the core plugin
        self.l3_plugin.auto_schedule_routers = (
            self.l3_plugin.auto_schedule_routers)
        super(L3RoutertypeAwareChanceL3AgentSchedulerTestCase,
              self).test_scheduler_auto_schedule_when_agent_added()

    @unittest.skip("DVR not supported")
    def test_get_l3_agent_candidates_dvr_ha_snat_no_vms(self):
        pass

    @unittest.skip("DVR not supported")
    def test_get_l3_agent_candidates_dvr_no_vms(self):
        pass

    @unittest.skip("DVR not supported")
    def test_add_router_to_l3_agent_dvr_to_snat(self):
        pass

    @unittest.skip("DVR not supported")
    def test_remove_router_from_l3_agent_in_dvr_snat_mode(self):
        pass

    @unittest.skip("DVR not supported")
    def test_add_router_to_l3_agent_mismatch_error_dvr_to_legacy(self):
        pass


class L3RoutertypeAwareLeastRoutersL3AgentSchedulerTestCase(
    test_l3_agent_scheduler.L3AgentLeastRoutersSchedulerTestCase,
        L3RoutertypeAwareL3AgentSchedulerTestCase):

    def setUp(self):
        cfg.CONF.set_override('router_scheduler_driver',
                              'neutron.scheduler.l3_agent_scheduler.'
                              'LeastRoutersScheduler')
        # call grandparent's setUp() to avoid that wrong scheduler is used
        super(test_l3_agent_scheduler.L3AgentLeastRoutersSchedulerTestCase,
              self).setUp()
        # Some UTs in parent class expects self.plugin to refer to l3 plugin
        self.plugin = self.l3_plugin

    @unittest.skip("DVR not supported")
    def test_add_router_to_l3_agent_mismatch_error_dvr_to_legacy(self):
        pass

    @unittest.skip("DVR not supported")
    def test_remove_router_from_l3_agent_in_dvr_mode(self):
        pass

    @unittest.skip("DVR not supported")
    def test_add_router_to_l3_agent_mismatch_error_legacy_to_dvr(self):
        pass

    @unittest.skip("DVR not supported")
    def test_add_router_to_l3_agent_dvr_to_snat(self):
        pass

    @unittest.skip("DVR not supported")
    def test_remove_router_from_l3_agent_in_dvr_snat_mode(self):
        pass

    @unittest.skip("DVR not supported")
    def test_get_l3_agent_candidates_dvr_snat(self):
        pass


#TODO(bobmel): Activate unit tests for DVR when we support DVR

class RouterHostingDeviceSchedulerTestMixIn(
        test_agentschedulers_db.AgentSchedulerTestMixIn):

    def _list_routers_hosted_by_hosting_device(self, hosting_device_id,
                                               expected_code=exc.HTTPOk.code,
                                               admin_context=True):
        path = "/dev_mgr/hosting_devices/%s/%s.%s" % (
            hosting_device_id, routertypeawarescheduler.DEVICE_L3_ROUTERS,
            self.fmt)
        return self._request_list(path, expected_code=expected_code,
                                  admin_context=admin_context)

    def _list_hosting_devices_hosting_router(self, router_id,
                                             expected_code=exc.HTTPOk.code,
                                             admin_context=True):
        path = "/routers/%s/%s.%s" % (
            router_id, routertypeawarescheduler.L3_ROUTER_DEVICES, self.fmt)
        return self._request_list(path, expected_code=expected_code,
                                  admin_context=admin_context)

    def _add_router_to_hosting_device(self, hosting_device_id, router_id,
                                      expected_code=exc.HTTPCreated.code,
                                      admin_context=True):
        path = "/dev_mgr/hosting_devices/%s/%s.%s" % (
            hosting_device_id, routertypeawarescheduler.DEVICE_L3_ROUTERS,
            self.fmt)
        req = self._path_create_request(path,
                                        {'router_id': router_id},
                                        admin_context=admin_context)
        res = req.get_response(self.ext_api)
        self.assertEqual(expected_code, res.status_int)

    def _remove_router_from_hosting_device(
            self, hosting_device_id, router_id,
            expected_code=exc.HTTPNoContent.code, admin_context=True):
        path = "/dev_mgr/hosting_devices/%s/%s/%s.%s" % (
            hosting_device_id, routertypeawarescheduler.DEVICE_L3_ROUTERS,
            router_id, self.fmt)
        req = self._path_delete_request(path, admin_context=admin_context)
        res = req.get_response(self.ext_api)
        self.assertEqual(expected_code, res.status_int)


class L3RoutertypeAwareHostingDeviceSchedulerTestCaseBase(
    test_l3.L3NatTestCaseMixin,
    RouterHostingDeviceSchedulerTestMixIn,
        test_l3_router_appliance_plugin.L3RouterApplianceTestCaseBase):

    router_type = 'ASR1k_Neutron_router'
    configure_routertypes = False

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None, use_ini_files=True):
        # save possible test_lib.test_config 'config_files' dict entry so we
        # can restore it after tests since we will change its value
        self._old_config_files = copy.copy(test_lib.test_config.get(
            'config_files'))
        if l3_plugin is None:
            l3_plugin = L3_PLUGIN_KLASS
        if ext_mgr is None:
            ext_mgr = TestSchedulingL3RouterApplianceExtensionManager()
        super(L3RoutertypeAwareHostingDeviceSchedulerTestCaseBase, self).setUp(
            core_plugin, l3_plugin, dm_plugin, ext_mgr)
        if use_ini_files is True:
            # include config files for device manager service plugin and router
            # service plugin since we define a number of hosting device
            # templates, hosting devices and routertypes there
            self._add_device_manager_plugin_ini_file()
            self._add_router_plugin_ini_file()
        #TODO(bobmel): Fix bug in test_extensions.py and we can remove the
        # below call to setup_config()
        self.setup_config()
        # do pool management in same green thread during tests
        self._mock_eventlet_greenpool_spawn_n()
        self.adminContext = bc.context.get_admin_context()
        # tests need a predictable random.choice so we always return first
        # item in the argument sequence
        self.random_patch = mock.patch('random.choice')
        random_mock = self.random_patch.start()

        random_mock.side_effect = lambda seq: sorted(seq)[0]

    def tearDown(self):
        self.random_patch.stop()
        if self._old_config_files is None:
            test_lib.test_config.pop('config_files', None)
        else:
            test_lib.test_config['config_files'] = self._old_config_files
        self._test_remove_all_hosting_devices()
        super(L3RoutertypeAwareHostingDeviceSchedulerTestCaseBase,
              self).tearDown()


class L3RoutertypeAwareHostingDeviceSchedulerBaseTestCase(
        L3RoutertypeAwareHostingDeviceSchedulerTestCaseBase):

    def test_new_router_backlogged_and_remains_backlogged_if_no_hosting_device(
            self):
        with mock.patch.object(self.l3_plugin, '_backlogged_routers') as m1,\
                mock.patch.object(self.l3_plugin, '_refresh_router_backlog',
                                  False):
            back_log = set()
            m1.__iter__ = lambda obj: iter(copy.deepcopy(back_log))
            m1.__contains__ = lambda obj, r_id: r_id in back_log
            m1.add.side_effect = lambda r_id: back_log.add(r_id)
            m1.discard.side_effect = lambda r_id: back_log.discard(r_id)
            arg_list = (routertype.TYPE_ATTR, )
            kwargs = {
                routertype.TYPE_ATTR: '00000000-0000-0000-0000-000000000007'}
            router = self._make_router(self.fmt, _uuid(), 'router1',
                                       arg_list=arg_list, **kwargs)
            r = router['router']
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            # verify that the new router is backlogged
            m1.add.assert_called_once_with(r['id'])
            self.assertIn(r['id'], back_log)
            self.l3_plugin._process_backlogged_routers()
            # verify that the router remains backlogged since not hosting
            # device exists for the router type
            self.assertIn(r['id'], back_log)

    def test_backlogged_router_is_scheduled_if_hosting_device_exists(self):
        with mock.patch.object(self.l3_plugin, '_backlogged_routers') as m1,\
                mock.patch.object(self.l3_plugin, '_refresh_router_backlog',
                                  False):
            back_log = set()
            # o'boy, this __iter__ mock took me so long to figure out....
            m1.__iter__ = lambda obj: iter(copy.deepcopy(back_log))
            m1.__contains__ = lambda obj, r_id: r_id in back_log
            m1.add.side_effect = lambda r_id: back_log.add(r_id)
            m1.discard.side_effect = lambda r_id: back_log.discard(r_id)
            arg_list = (routertype.TYPE_ATTR, )
            kwargs = {
                routertype.TYPE_ATTR: '00000000-0000-0000-0000-000000000006'}
            router = self._make_router(self.fmt, _uuid(), 'router1',
                                       arg_list=arg_list, **kwargs)
            r = router['router']
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            m1.add.assert_called_once_with(r['id'])
            self.l3_plugin._process_backlogged_routers()
            self.assertNotIn(r['id'], back_log)

    def test_already_backlogged_router_not_backlogged(self):
        with mock.patch.object(self.l3_plugin, '_backlogged_routers') as m:
            back_log = set()
            m.__contains__.side_effect = lambda r_id: r_id in back_log
            m.add.side_effect = lambda r_id: back_log.add(r_id)
            with self.router() as router:
                r = router['router']
                self.assertIsNone(r[HOSTING_DEVICE_ATTR])
                m.add.assert_called_once_with(r['id'])
                r_after = self._update(
                    'routers', r['id'],
                    {'router': {'name': 'routerA'}})['router']
                self.assertEqual('routerA', r_after['name'])
                self.assertIsNone(
                    r_after[HOSTING_DEVICE_ATTR])
                # router should be back-logged only once for later
                # scheduling attempts
                m.add.assert_called_once_with(r['id'])
                m.__contains__.assert_has_calls([mock.call(r['id']),
                                                 mock.call(r['id'])])

    def test_namespace_router_not_backlogged(self):
        with mock.patch.object(self.l3_plugin, '_backlogged_routers') as m:
            back_log = set()
            m.__contains__.side_effect = lambda r_id: r_id in back_log
            m.add.side_effect = lambda r_id: back_log.add(r_id)
            arg_list = (routertype.TYPE_ATTR, )
            kwargs = {
                routertype.TYPE_ATTR: '00000000-0000-0000-0000-000000000001'}
            router = self._make_router(self.fmt, _uuid(), 'router1',
                                       arg_list=arg_list, **kwargs)
            r = router['router']
            self.assertFalse(m.add.called)
            self.assertFalse(m.__contains__.called)
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            r_after = self._update('routers', r['id'],
                                   {'router': {'name': 'routerA'}})['router']
            self.assertEqual('routerA', r_after['name'])
            self.assertIsNone(
                r_after[HOSTING_DEVICE_ATTR])
            # router should be back-logged only once for later
            # scheduling attempts
            self.assertFalse(m.add.called)
            self.assertFalse(m.__contains__.called)

    def test_router_without_auto_schedule_not_backlogged(self):
        with mock.patch.object(self.l3_plugin, '_backlog_router') as (
                mock_b_lg):
            arg_list = (routertypeawarescheduler.AUTO_SCHEDULE_ATTR, )
            kwargs = {
                routertypeawarescheduler.AUTO_SCHEDULE_ATTR: False}
            router = self._make_router(self.fmt, _uuid(), 'router1',
                                       arg_list=arg_list, **kwargs)
            r = router['router']
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            mock_b_lg.assert_has_calls([])
            r_after = self._update('routers', r['id'],
                                   {'router': {'name': 'routerA'}})['router']
            self.assertIsNone(r_after[HOSTING_DEVICE_ATTR])
            mock_b_lg.assert_has_calls([])
            self.l3_plugin._sync_router_backlog()
            self.assertEqual(0, len(self.l3_plugin._backlogged_routers))

    def test_rpc_sync_routers_ext_gets_no_namespace_routers(self):
        arg_list = (routertype.TYPE_ATTR, )
        kwargs = {routertype.TYPE_ATTR: test_db_routertype.NS_ROUTERTYPE_NAME}
        # namespace-based routers
        self._make_router(self.fmt, _uuid(), 'router1', arg_list=arg_list,
                          **kwargs)
        # hw routers
        with self.router(name='router2') as r2, self.router(
                name='router3') as r3:
            # router 4
            self._make_router(self.fmt, _uuid(), 'router4', arg_list=arg_list,
                              **kwargs)
            routers = self.l3_plugin.get_sync_data_ext(self.adminContext)
            r_ids = set(r['id'] for r in routers)
            self.assertEqual(2, len(r_ids))
            for r in [r2['router'], r3['router']]:
                self.assertIn(r['id'], r_ids)

    def test_router_scheduling_aborts_if_other_process_scheduled_router(self):

        def fake_allocator(context, binding_info_db, target_hosting_device_id,
                           slot_need, synchronized):
            # here we mimic that another process has concurrently already
            # completed the scheduling of the router so that the attempt here
            # fails and raises an IntegrityError
            res = orig_func(context, binding_info_db, selected_hd_id,
                            slot_need, synchronized)
            self.assertTrue(res)
            self.assertEqual(0, len(self.l3_plugin._backlogged_routers))
            # The call to orig_func removed the router from the one and only
            # backlog we have in the test when it bound the router to the
            # hosting device.
            # However, in the real, non-simulated case each process would have
            # its own backlog. We therefore put the router back in the backlog
            # to mimic this.
            self.l3_plugin._backlogged_routers.add(binding_info_db.router_id)
            # kaboom!
            raise db_exc.DBDuplicateEntry(
                inner_exception=inner_db_exc.IntegrityError(
                    "Invalid insert", params="", orig=None))

        selected_hd_id = '00000000-0000-0000-0000-000000000002'
        with self.router() as router:
            r = router['router']
            self.assertIn(r['id'], self.l3_plugin._backlogged_routers)
            orig_func = self.l3_plugin._try_allocate_slots_and_bind_to_host
            with mock.patch.object(
                    self.l3_plugin,
                    '_try_allocate_slots_and_bind_to_host') as m:
                m.side_effect = fake_allocator
                self.l3_plugin._process_backlogged_routers()
            self.assertEqual(1, len(self.l3_plugin._backlogged_routers))
            r_after = self._show('routers', r['id'])['router']
            self.assertEqual(selected_hd_id, r_after[HOSTING_DEVICE_ATTR])

    def test_router_deleted_by_other_process_removed_from_backlog(self):
        r_id = 'non_existant_router_id'
        self.l3_plugin._backlogged_routers.add(r_id)
        self.l3_plugin._refresh_router_backlog = False
        self.assertEqual(1, len(self.l3_plugin._backlogged_routers))
        self.l3_plugin._process_backlogged_routers()
        self.assertEqual(0, len(self.l3_plugin._backlogged_routers))


class L3RoutertypeAwareHostingDeviceSchedulerTestCase(
        L3RoutertypeAwareHostingDeviceSchedulerTestCaseBase):

    def test_router_add_to_hosting_device(self):
        with self.router() as router:
            r = router['router']
            rt_id = '00000000-0000-0000-0000-000000000005'
            self.assertEqual(rt_id, r[routertype.TYPE_ATTR])
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            self._add_router_to_hosting_device(
                '00000000-0000-0000-0000-000000000001', r['id'])
            r_after = self._show('routers', r['id'])['router']
            self.assertEqual(rt_id, r_after[routertype.TYPE_ATTR])
            self.assertEqual('00000000-0000-0000-0000-000000000001',
                             r_after[HOSTING_DEVICE_ATTR])

    def test_hosted_router_add_to_hosting_device(self):
        with self.router() as router:
            # trigger scheduling of router
            self.l3_plugin._process_backlogged_routers()
            r = self._show('routers', router['router']['id'])['router']
            rt_id = '00000000-0000-0000-0000-000000000005'
            self.assertEqual(rt_id, r[routertype.TYPE_ATTR])
            self.assertIsNotNone(r[HOSTING_DEVICE_ATTR])
            self._add_router_to_hosting_device(
                '00000000-0000-0000-0000-000000000002', r['id'],
                exc.HTTPConflict.code)
            r_after = self._show('routers', r['id'])['router']
            self.assertEqual(rt_id, r_after[routertype.TYPE_ATTR])
            self.assertEqual(r[HOSTING_DEVICE_ATTR],
                             r_after[HOSTING_DEVICE_ATTR])

    def test_hosted_router_add_to_different_type_hosting_device(self):
        with self.router() as router:
            r = router['router']
            rt_id = '00000000-0000-0000-0000-000000000005'
            self.assertEqual(rt_id, r[routertype.TYPE_ATTR])
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            hd_id = '00000000-0000-0000-0000-000000000004'
            self._add_router_to_hosting_device(hd_id, r['id'])
            r_after = self._show('routers', r['id'])['router']
            self.assertEqual(hd_id, r_after[HOSTING_DEVICE_ATTR])
            temp_rt_id = '00000000-0000-0000-0000-000000000006'
            routertype_id = "%s (normal: %s)" % (temp_rt_id, rt_id)
            self.assertEqual(routertype_id, r_after[routertype.TYPE_ATTR])

    def test_router_add_to_hosting_device_insufficient_slots(self):
        with mock.patch.object(
                self.core_plugin,
                'acquire_hosting_device_slots') as acquire_mock,\
                self.router() as router:
            acquire_mock.return_value = False
            r = router['router']
            rt_id = '00000000-0000-0000-0000-000000000005'
            self.assertEqual(rt_id, r[routertype.TYPE_ATTR])
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            hd_id = '00000000-0000-0000-0000-000000000004'
            self._add_router_to_hosting_device(hd_id, r['id'],
                                               exc.HTTPConflict.code)
            r_after = self._show('routers', r['id'])['router']
            self.assertIsNone(r_after[HOSTING_DEVICE_ATTR])
            self.assertEqual(rt_id, r_after[routertype.TYPE_ATTR])

    def test_router_add_to_hosting_device_insufficient_slots_no_auto(self):
        with mock.patch.object(
                self.core_plugin,
                'acquire_hosting_device_slots') as acquire_mock,\
                mock.patch.object(self.l3_plugin,
                                  '_backlog_router') as mock_b_lg:
            acquire_mock.return_value = False
            arg_list = (routertypeawarescheduler.AUTO_SCHEDULE_ATTR, )
            kwargs = {
                routertypeawarescheduler.AUTO_SCHEDULE_ATTR: False}
            router = self._make_router(self.fmt, _uuid(), 'router1',
                                       arg_list=arg_list, **kwargs)
            r = router['router']
            rt_id = '00000000-0000-0000-0000-000000000005'
            self.assertEqual(rt_id, r[routertype.TYPE_ATTR])
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            mock_b_lg.assert_has_calls([])
            hd_id = '00000000-0000-0000-0000-000000000004'
            self._add_router_to_hosting_device(hd_id, r['id'],
                                               exc.HTTPConflict.code)
            r_after = self._show('routers', r['id'])['router']
            self.assertIsNone(r_after[HOSTING_DEVICE_ATTR])
            self.assertEqual(rt_id, r_after[routertype.TYPE_ATTR])
            mock_b_lg.assert_has_calls([])

    def test_router_add_to_hosting_device_with_admin_state_down(self):
        with self.router() as router:
            r = router['router']
            rt_id = '00000000-0000-0000-0000-000000000005'
            self.assertEqual(rt_id, r[routertype.TYPE_ATTR])
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            id_hd_disabled = '00000000-0000-0000-0000-000000000001'
            self._update('hosting_devices', id_hd_disabled,
                         {'hosting_device': {'admin_state_up': False}})
            self._add_router_to_hosting_device(id_hd_disabled, r['id'],
                                               exc.HTTPNotFound.code)
            r_after = self._show('routers', r['id'])['router']
            self.assertEqual(rt_id, r_after[routertype.TYPE_ATTR])
            self.assertIsNone(r_after[HOSTING_DEVICE_ATTR])

    def test_router_add_to_hosting_device_two_times(self):
        with self.router() as router:
            r = router['router']
            rt_id = '00000000-0000-0000-0000-000000000005'
            self.assertEqual(rt_id, r[routertype.TYPE_ATTR])
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            self._add_router_to_hosting_device(
                '00000000-0000-0000-0000-000000000001', r['id'])
            r_after = self._show('routers', r['id'])['router']
            self.assertEqual(rt_id, r_after[routertype.TYPE_ATTR])
            self.assertEqual('00000000-0000-0000-0000-000000000001',
                             r_after[HOSTING_DEVICE_ATTR])
            self._add_router_to_hosting_device(
                '00000000-0000-0000-0000-000000000001', r['id'])
            r_final = self._show('routers', r['id'])['router']
            self.assertEqual(rt_id, r_final[routertype.TYPE_ATTR])
            self.assertEqual('00000000-0000-0000-0000-000000000001',
                             r_final[HOSTING_DEVICE_ATTR])

    def test_router_remove_from_hosting_device(self):
        with self.router() as router:
            # trigger scheduling of router
            self.l3_plugin._process_backlogged_routers()
            r = self._show('routers', router['router']['id'])['router']
            rt_id = '00000000-0000-0000-0000-000000000005'
            self.assertEqual(rt_id, r[routertype.TYPE_ATTR])
            self.assertIsNotNone(r[HOSTING_DEVICE_ATTR])
            self._remove_router_from_hosting_device(
                '00000000-0000-0000-0000-000000000001', r['id'])
            r_after = self._show('routers', r['id'])['router']
            self.assertEqual(rt_id, r_after[routertype.TYPE_ATTR])
            self.assertIsNone(r_after[HOSTING_DEVICE_ATTR])

    def test_router_remove_from_wrong_hosting_device(self):
        with self.router() as router:
            # trigger scheduling of router
            self.l3_plugin._process_backlogged_routers()
            r = self._show('routers', router['router']['id'])['router']
            rt_id = '00000000-0000-0000-0000-000000000005'
            self.assertEqual(rt_id, r[routertype.TYPE_ATTR])
            self.assertIsNotNone(r[HOSTING_DEVICE_ATTR])
            self._remove_router_from_hosting_device(
                '00000000-0000-0000-0000-000000000002', r['id'],
                exc.HTTPConflict.code)
            r_after = self._show('routers', r['id'])['router']
            self.assertEqual(rt_id, r_after[routertype.TYPE_ATTR])
            self.assertEqual(r[HOSTING_DEVICE_ATTR],
                             r_after[HOSTING_DEVICE_ATTR])

    def test_unhosted_router_remove_from_hosting_device(self):
        with self.router() as router:
            r = router['router']
            rt_id = '00000000-0000-0000-0000-000000000005'
            self.assertEqual(rt_id, r[routertype.TYPE_ATTR])
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            self._remove_router_from_hosting_device(
                '00000000-0000-0000-0000-000000000001', r['id'],
                exc.HTTPConflict.code)
            r_after = self._show('routers', r['id'])['router']
            self.assertEqual(rt_id, r_after[routertype.TYPE_ATTR])
            self.assertIsNone(r_after[HOSTING_DEVICE_ATTR])

    def test_router_scheduling_policy(self):
        with self.router() as router1, self.router() as router2:
            r1 = router1['router']
            r2 = router2['router']
            hd_id = '00000000-0000-0000-0000-000000000001'
            self._add_router_to_hosting_device(hd_id, r1['id'])
            self._list_routers_hosted_by_hosting_device(hd_id)
            self._list_routers_hosted_by_hosting_device(
                hd_id, expected_code=exc.HTTPForbidden.code,
                admin_context=False)
            self._add_router_to_hosting_device(
                hd_id, r2['id'], expected_code=exc.HTTPForbidden.code,
                admin_context=False)
            self._remove_router_from_hosting_device(
                hd_id, r1['id'], expected_code=exc.HTTPForbidden.code,
                admin_context=False)
            self._remove_router_from_hosting_device(hd_id, r1['id'])
            self._list_hosting_devices_hosting_router(r1['id'])
            self._list_hosting_devices_hosting_router(
                r1['id'], expected_code=exc.HTTPForbidden.code,
                admin_context=False)

    def test_hosting_device_keep_services_off(self):
        #TODO(bobmel): Implement this unit test
        # Introduce new option: keep_services_on_agents_with_admin_state_down
        # Here set to keep_services_on_agents_with_admin_state_down = False
        # routers on hosting device that is set to admin down should be removed
        #  from that hosting device
        pass

    def test_hosting_device_keep_services_on(self):
        #TODO(bobmel): Implement this unit test
        # Introduce new option: keep_services_on_agents_with_admin_state_down
        # Here set to keep_services_on_agents_with_admin_state_down = False
        # routers on hosting device that set to admin down should stay on that
        # hosting device
        pass

    def test_list_routers_by_hosting_device(self):
        with self.router() as router1, self.router() as router2,\
                self.router() as router3:
            r1 = router1['router']
            r2 = router2['router']
            r3 = router3['router']
            hd1_id = '00000000-0000-0000-0000-000000000001'
            self._add_router_to_hosting_device(hd1_id, r1['id'])
            self._add_router_to_hosting_device(hd1_id, r2['id'])
            hd2_id = '00000000-0000-0000-0000-000000000002'
            self._add_router_to_hosting_device(hd2_id, r3['id'])
            r_list1 = self._list_routers_hosted_by_hosting_device(hd1_id)
            self.assertEqual(2, len(r_list1['routers']))
            r1_set = {r1['id'], r2['id']}
            for r in r_list1['routers']:
                self.assertTrue(r1_set)
            r_list2 = self._list_routers_hosted_by_hosting_device(hd2_id)
            self.assertEqual(1, len(r_list2['routers']))
            self.assertEqual(r3['id'], r_list2['routers'][0]['id'])

    def test_list_routers_by_hosting_device_with_non_existing_hosting_device(
            self):
        with self.router() as router:
            r = router['router']
            hd_id = '00000000-0000-0000-0000-000000000001'
            missing_hd_id = '00000000-0000-0000-0000-000000000099'
            self._add_router_to_hosting_device(hd_id, r['id'])
            r_list = self._list_routers_hosted_by_hosting_device(missing_hd_id)
            self.assertEqual(0, len(r_list['routers']))

    def test_list_hosting_devices_hosting_router(self):
        with self.router() as router:
            r = router['router']
            hd_id = '00000000-0000-0000-0000-000000000001'
            self._add_router_to_hosting_device(hd_id, r['id'])
            h_list = self._list_hosting_devices_hosting_router(r['id'])
            self.assertEqual(1, len(h_list['hosting_devices']))
            self.assertEqual(hd_id, h_list['hosting_devices'][0]['id'])

    def test_list_hosting_devices_hosting_unhosted_router(self):
        with self.router() as router:
            r = router['router']
            h_list = self._list_hosting_devices_hosting_router(r['id'])
            self.assertEqual(0, len(h_list['hosting_devices']))

    def test_list_hosting_devices_hosting_non_existent_router(self):
        with self.router() as router:
            r = router['router']
            r_id_non_exist = r['id'][:-1]
            h_list = self._list_hosting_devices_hosting_router(r_id_non_exist)
            self.assertEqual(0, len(h_list['hosting_devices']))

    def _test_list_active_sync_routers_on_hosting_devices(self, func):
        with self.router(name='router1') as router1,\
                self.router(name='router2') as router2,\
                self.router(name='router3') as router3:
            r1 = router1['router']
            r2 = router2['router']
            r3 = router3['router']
            hd1_id = '00000000-0000-0000-0000-000000000001'
            self._add_router_to_hosting_device(hd1_id, r1['id'])
            self._add_router_to_hosting_device(hd1_id, r2['id'])
            hd2_id = '00000000-0000-0000-0000-000000000002'
            self._add_router_to_hosting_device(hd2_id, r3['id'])
            # when cfg agent on host_a registers itself, hosting devices with
            # uuid hd1_id and hd2_id will be assigned to that cfg agent
            self._register_cfg_agent_states()
            template_id = '00000000-0000-0000-0000-000000000005'
            with self.hosting_device(template_id, no_delete=True) as h_d3,\
                    self.router(name='router4') as router4,\
                    self.router(name='router5') as router5:
                # hd3 should not yet have been assigned to a cfg agent
                hd3 = h_d3['hosting_device']
                r4 = router4['router']
                r5 = router5['router']
                self._add_router_to_hosting_device(hd3['id'], r4['id'])
                self._add_router_to_hosting_device(hd3['id'], r5['id'])
                # when cfg agent on host_b registers itself, hosting
                # device hd3 will be assigned to that cfg agent
                self._register_cfg_agent_states(host_a_active=False,
                                                host_b_active=True)
                agents = self._list(
                    'agents',
                    query_params='agent_type=%s' % c_const.AGENT_TYPE_CFG)
                agent_dict = {agt['host']: agt for agt in agents['agents']}
                if func:
                    func(r1, r2, r3, r4, r5, hd1_id, hd2_id, hd3,
                         template_id, agent_dict)

    def _verify_hosting(self, agent_dict, r_set, host, router_ids=None,
                        hosting_device_ids=None):
        r_l = self.l3_plugin.list_active_sync_routers_on_hosting_devices(
            self.adminContext, host, router_ids, hosting_device_ids)
        self.assertEqual(len(r_set), len(r_l))
        hds = {}
        agent_id = agent_dict[host]['id']
        for r in r_l:
            self.assertTrue(r['id'] in r_set)
            router_host = r[HOSTING_DEVICE_ATTR]
            if router_host not in hds:
                hd = self._show('hosting_devices', router_host)
                hds[router_host] = hd['hosting_device']
            self.assertEqual(agent_id, hds[router_host]['cfg_agent_id'])

    def test_list_active_sync_all_routers_on_all_hosting_devices(self):

        def assert_function(r1, r2, r3, r4, r5, hd1_id, hd2_id,
                            hd3, template_id, agent_dict):
            r_set = {r1['id'], r2['id'], r3['id']}
            self._verify_hosting(agent_dict, r_set,
                                 device_manager_test_support.L3_CFG_HOST_A)
            r_set = {r4['id'], r5['id']}
            self._verify_hosting(agent_dict, r_set,
                                 device_manager_test_support.L3_CFG_HOST_B)

        self._test_list_active_sync_routers_on_hosting_devices(assert_function)

    def test_list_active_sync_some_routers_on_all_hosting_devices(self):

        def assert_function(r1, r2, r3, r4, r5, hd1_id, hd2_id,
                            hd3, template_id, agent_dict):
            router_ids = [r1['id'], r3['id']]
            r_set = set(router_ids)
            self._verify_hosting(agent_dict, r_set,
                                 device_manager_test_support.L3_CFG_HOST_A,
                                 router_ids)

        self._test_list_active_sync_routers_on_hosting_devices(assert_function)

    def test_list_active_sync_all_routers_on_some_hosting_devices(self):
        def assert_function(r1, r2, r3, r4, r5, hd1_id, hd2_id,
                            hd3, template_id, agent_dict):
            r_set = {r3['id']}
            self._verify_hosting(agent_dict, r_set,
                                 device_manager_test_support.L3_CFG_HOST_A,
                                 hosting_device_ids=[hd2_id])

        self._test_list_active_sync_routers_on_hosting_devices(assert_function)

    def test_list_active_sync_some_routers_on_some_hosting_devices(self):
        def assert_function(r1, r2, r3, r4, r5, hd1_id, hd2_id,
                            hd3, template_id, agent_dict):
            router_ids = [r1['id'], r3['id']]
            r_set = {r1['id']}
            self._verify_hosting(agent_dict, r_set,
                                 device_manager_test_support.L3_CFG_HOST_A,
                                 router_ids, hosting_device_ids=[hd1_id])

        self._test_list_active_sync_routers_on_hosting_devices(assert_function)

    def test_list_active_sync_routers_on_hosting_devices_cfg_agent_admin_down(
            self):

        def assert_function(r1, r2, r3, r4, r5, hd1_id, hd2_id,
                            hd3, template_id, agent_dict):
            agent_id = agent_dict[device_manager_test_support.L3_CFG_HOST_A][
                'id']
            self._update('agents', agent_id,
                         {'agent': {'admin_state_up': False}})
            r_set = {}
            self._verify_hosting(agent_dict, r_set,
                                 device_manager_test_support.L3_CFG_HOST_A)
            r_set = {r4['id'], r5['id']}
            self._verify_hosting(agent_dict, r_set,
                                 device_manager_test_support.L3_CFG_HOST_B)

        self._test_list_active_sync_routers_on_hosting_devices(assert_function)

    def test_list_active_sync_routers_on_hosting_devices_idle_cfg_agent(
            self):

        def assert_function(r1, r2, r3, r4, r5, hd1_id, hd2_id,
                            hd3, template_id, agent_dict):
            # there should be no hosting devices left that can be assigned
            # to cfg agent on host_c
            self._register_cfg_agent_states(host_a_active=False,
                                            host_b_active=False,
                                            host_c_active=True)
            r_set = {}
            agents = self._list(
                'agents',
                query_params='agent_type=%s' % c_const.AGENT_TYPE_CFG)
            agent_dict = {agt['host']: agt for agt in agents['agents']}
            self._verify_hosting(agent_dict, r_set,
                                 device_manager_test_support.L3_CFG_HOST_C)
            r_set = {r1['id'], r2['id'], r3['id']}
            self._verify_hosting(agent_dict, r_set,
                                 device_manager_test_support.L3_CFG_HOST_A)
            r_set = {r4['id'], r5['id']}
            self._verify_hosting(agent_dict, r_set,
                                 device_manager_test_support.L3_CFG_HOST_B)

        self._test_list_active_sync_routers_on_hosting_devices(assert_function)

    def test_list_active_sync_routers_on_hosting_devices_no_cfg_agent_on_host(
            self):
        self.assertRaises(
            agent_exceptions.AgentNotFoundByTypeHost,
            self.l3_plugin.list_active_sync_routers_on_hosting_devices,
            self.adminContext, 'bogus_host')

    def test_list_all_routers_on_hosting_devices(self):
        with self.router(name='router1') as router1,\
                self.router(name='router2'),\
                self.router(name='router3') as router3:
            r1 = router1['router']
            r3 = router3['router']
            hd1_id = '00000000-0000-0000-0000-000000000001'
            self._add_router_to_hosting_device(hd1_id, r1['id'])
            hd2_id = '00000000-0000-0000-0000-000000000002'
            self._add_router_to_hosting_device(hd2_id, r3['id'])
            template_id = '00000000-0000-0000-0000-000000000005'
            with self.hosting_device(template_id, no_delete=True) as h_d3,\
                    self.router(name='router4') as router4,\
                    self.router(name='router5') as router5:
                hd3 = h_d3['hosting_device']
                r4 = router4['router']
                r5 = router5['router']
                self._add_router_to_hosting_device(hd3['id'], r4['id'])
                self._add_router_to_hosting_device(hd3['id'], r5['id'])
                r_l = self.l3_plugin.list_all_routers_on_hosting_devices(
                    self.adminContext)
                r_set = {r1['id'], r3['id'], r4['id'], r5['id']}
                self.assertEqual(len(r_set), len(r_l))
                hds = {}
                for r in r_l:
                    self.assertTrue(r['id'] in r_set)
                    router_host = r[HOSTING_DEVICE_ATTR]
                    self.assertIsNotNone(router_host)
                    # assert that router's hosting device exists
                    if router_host not in hds:
                        hd = self._show('hosting_devices', router_host)
                        hds[router_host] = hd['hosting_device']

    def test_router_reschedule_from_dead_hosting_device(self):
        with mock.patch.object(self.l3_plugin, '_backlogged_routers') as m1,\
                mock.patch.object(self.l3_plugin, '_refresh_router_backlog',
                                  False):
            back_log = set()
            m1.__iter__ = lambda obj: iter(copy.deepcopy(back_log))
            m1.__contains__ = lambda obj, r_id: r_id in back_log
            m1.add.side_effect = lambda r_id: back_log.add(r_id)
            m1.discard.side_effect = lambda r_id: back_log.discard(r_id)
            with self.router(name='router1') as router1,\
                    self.router(name='router2') as router2,\
                    self.router(name='router3') as router3:
                rs_initial = [r['router'] for r in [router1, router2, router3]]
                # all routers initially un-hosted
                for r in rs_initial:
                    self.assertIsNone(r[HOSTING_DEVICE_ATTR])
                hosting_device_id1 = '00000000-0000-0000-0000-000000000001'
                hosting_device_id2 = '00000000-0000-0000-0000-000000000002'
                r_ids = [r['id'] for r in [rs_initial[0], rs_initial[2]]]
                for r_id in r_ids:
                    self._add_router_to_hosting_device(hosting_device_id1,
                                                       r_id)
                self._add_router_to_hosting_device(hosting_device_id2,
                                                   rs_initial[1]['id'])
                rs_after = [self._show('routers', r['id'])['router']
                            for r in rs_initial]
                # r1 and r3 on hosting device 1, r2 on hosting device 2
                for r in [rs_after[0], rs_after[2]]:
                    self.assertEqual(hosting_device_id1,
                                     r[HOSTING_DEVICE_ATTR])
                self.assertEqual(hosting_device_id2,
                                 rs_after[1][HOSTING_DEVICE_ATTR])
                # no routers should be back-logged now
                self.assertEqual(0, len(back_log))
                hosting_device_1 = self._show(
                    'hosting_devices', hosting_device_id1)['hosting_device']
                affected_resources = {}
                notifier_mock = mock.MagicMock()
                with mock.patch.dict(self.l3_plugin.agent_notifiers,
                                     {AGENT_TYPE_L3_CFG: notifier_mock}):
                    # now report hosting device 1 as dead
                    self.l3_plugin.handle_non_responding_hosting_devices(
                        self.adminContext, [hosting_device_1],
                        affected_resources)
                # only routers 1 and 3 should be affected
                affected_rs = affected_resources[hosting_device_id1]['routers']
                self.assertEqual(2, len(affected_rs))
                ntfy_method = notifier_mock.routers_removed_from_hosting_device
                ntfy_method.assert_called_with(mock.ANY, affected_rs,
                                               hosting_device_1)
                # affected routers should be back-logged
                for r_id in r_ids:
                    self.assertIn(r_id, affected_rs)
                    self.assertIn(r_id, back_log)
                rs_final = [self._show('routers', r['id'])['router']
                            for r in rs_initial]
                # routers 1 and 3 should be un-hosted
                for r in [rs_final[0], rs_final[2]]:
                    self.assertIsNone(r[HOSTING_DEVICE_ATTR])
                # router 2 was unaffected and should remain hosted
                self.assertEqual(hosting_device_id2,
                                 rs_final[1][HOSTING_DEVICE_ATTR])

    def test_router_without_auto_schedule_not_unscheduled_from_dead_hd(self):
        with mock.patch.object(
                self.l3_plugin, 'unschedule_router_from_hosting_device') as (
                    mock_unsched):
            arg_list = (routertypeawarescheduler.AUTO_SCHEDULE_ATTR, )
            kwargs = {
                routertypeawarescheduler.AUTO_SCHEDULE_ATTR: False}
            router = self._make_router(self.fmt, _uuid(), 'router1',
                                       arg_list=arg_list, **kwargs)
            r = router['router']
            r_id = r['id']
            self.assertIsNone(r[HOSTING_DEVICE_ATTR])
            hosting_device_id = '00000000-0000-0000-0000-000000000001'
            self._add_router_to_hosting_device(hosting_device_id, r_id)
            r_after = self._show('routers', r['id'])['router']
            self.assertEqual(hosting_device_id, r_after[HOSTING_DEVICE_ATTR])
            affected_resources = {}
            # now report hosting device 1 as dead
            self.l3_plugin.handle_non_responding_hosting_devices(
                self.adminContext, [{'id': hosting_device_id}],
                affected_resources)
            self.assertEqual(0, mock_unsched.call_count)
            r_final = self._show('routers', r['id'])['router']
            self.assertEqual(hosting_device_id, r_final[HOSTING_DEVICE_ATTR])


class HostingDeviceRouterL3CfgAgentNotifierTestCase(
        L3RoutertypeAwareHostingDeviceSchedulerTestCaseBase):

    mock_cfg_agent_notifiers = False

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None):
        super(HostingDeviceRouterL3CfgAgentNotifierTestCase, self).setUp(
            core_plugin, l3_plugin, dm_plugin, ext_mgr)
        fake_notifier.reset()

    def test_router_add_to_hosting_device_notification(self):
        l3_notifier = self.l3_plugin.agent_notifiers[c_const.AGENT_TYPE_L3_CFG]
        with mock.patch.object(
                l3_notifier.client, 'prepare',
                return_value=l3_notifier.client) as mock_prepare,\
                mock.patch.object(l3_notifier.client, 'cast') as mock_cast,\
                self.router() as router:
            r = router['router']
            # when cfg agent on host_a registers itself, hosting
            # device with uuid hd_id will be assigned to that cfg agent
            self._register_cfg_agent_states()
            self._add_router_to_hosting_device(
                '00000000-0000-0000-0000-000000000001', r['id'])
            mock_prepare.assert_called_with(
                server=device_manager_test_support.L3_CFG_HOST_A)
            mock_cast.assert_called_with(
                mock.ANY, 'router_added_to_hosting_device',
                routers=[r['id']])
            notifications = fake_notifier.NOTIFICATIONS
            expected_event_type = 'hosting_device.router.add'
            self._assert_notify(notifications, expected_event_type)

    def test_router_remove_from_hosting_device_notification(self):
        l3_notifier = self.l3_plugin.agent_notifiers[c_const.AGENT_TYPE_L3_CFG]
        with mock.patch.object(
                l3_notifier.client, 'prepare',
                return_value=l3_notifier.client) as mock_prepare,\
                mock.patch.object(l3_notifier.client, 'cast') as mock_cast,\
                self.router() as router:
            r = router['router']
            # when cfg agent on host_a registers itself, hosting
            # device with uuid hd_id will be assigned to that cfg agent
            self._register_cfg_agent_states()
            self._add_router_to_hosting_device(
                '00000000-0000-0000-0000-000000000001', r['id'])
            self._remove_router_from_hosting_device(
                '00000000-0000-0000-0000-000000000001', r['id'])
            mock_prepare.assert_called_with(
                server=device_manager_test_support.L3_CFG_HOST_A)
            mock_cast.assert_called_with(
                mock.ANY, 'router_removed_from_hosting_device',
                routers=[r['id']])
            notifications = fake_notifier.NOTIFICATIONS
            expected_event_type = 'hosting_device.router.remove'
            self._assert_notify(notifications, expected_event_type)

    def test_backlogged_routers_scheduled_routers_updated_notification(self):
        l3_notifier = self.l3_plugin.agent_notifiers[c_const.AGENT_TYPE_L3_CFG]
        with mock.patch.object(
                l3_notifier.client, 'prepare',
                return_value=l3_notifier.client) as mock_prepare,\
                mock.patch.object(l3_notifier.client, 'cast') as mock_cast,\
                mock.patch.object(self.l3_plugin,
                                  '_backlogged_routers') as mock_b_lg,\
                mock.patch.object(self.l3_plugin, '_refresh_router_backlog',
                                  False):
            back_log = set()
            mock_b_lg.__iter__ = lambda obj: iter(copy.deepcopy(back_log))
            mock_b_lg.__contains__ = lambda obj, r_id: r_id in back_log
            mock_b_lg.add.side_effect = lambda r_id: back_log.add(r_id)
            mock_b_lg.discard.side_effect = lambda r_id: back_log.discard(r_id)
            arg_list = (routertype.TYPE_ATTR, )
            # namespace-based router
            kwargs = {
                routertype.TYPE_ATTR: '00000000-0000-0000-0000-000000000001'}
            self._make_router(self.fmt, _uuid(), 'router1', arg_list=arg_list,
                              **kwargs)
            # router that should be successfully hosted
            kwargs = {
                routertype.TYPE_ATTR: '00000000-0000-0000-0000-000000000006'}
            r2 = self._make_router(self.fmt, _uuid(), 'router2',
                                   arg_list=arg_list, **kwargs)['router']
            # routertype for which no hosting devices exists
            kwargs = {
                routertype.TYPE_ATTR: '00000000-0000-0000-0000-000000000007'}
            r3 = self._make_router(self.fmt, _uuid(), 'router3',
                                   arg_list=arg_list, **kwargs)['router']
            # routers that should be successfully hosted
            with self.router(name='router4') as router4, self.router(
                    name='router5') as router5:
                # when cfg agent on host_a registers itself, hosting
                # devices will be assigned to that cfg agent
                self._register_cfg_agent_states()
                r4 = router4['router']
                r5 = router5['router']
                self.assertEqual(4, len(back_log))
                for r_id in [r2['id'], r3['id'], r4['id'], r5['id']]:
                    self.assertIn(r_id, back_log)
                self.l3_plugin._process_backlogged_routers()
                mock_prepare.assert_called_with(
                    server=device_manager_test_support.L3_CFG_HOST_A)
                calls = [
                    mock.call(mock.ANY, 'routers_updated', routers=[r2['id']]),
                    mock.call(mock.ANY, 'routers_updated', routers=[r4['id']]),
                    mock.call(mock.ANY, 'routers_updated', routers=[r5['id']])]
                mock_cast.assert_has_calls(calls, any_order=True)
                self.assertEqual(1, len(back_log))
                self.assertIn(r3['id'], back_log)


class TestHASchedulingL3RouterApplianceExtensionManager(
        TestSchedulingL3RouterApplianceExtensionManager):

    def get_resources(self):
        # add ha attributes to router resource
        l3_const.RESOURCE_ATTRIBUTE_MAP['routers'].update(
            ha.EXTENDED_ATTRIBUTES_2_0['routers'])
        # let our super class do the rest
        return super(TestHASchedulingL3RouterApplianceExtensionManager,
                     self).get_resources()


# A scheduler-enabled routertype capable L3 routing service plugin class
class TestSchedulingHACapableL3RouterServicePlugin(
        ha_db.HA_db_mixin, TestSchedulingCapableL3RouterServicePlugin):

    supported_extension_aliases = (
        TestSchedulingCapableL3RouterServicePlugin.
        supported_extension_aliases +
        [routertypeawarescheduler.ROUTERTYPE_AWARE_SCHEDULER_ALIAS,
         bc.constants.L3_AGENT_SCHEDULER_EXT_ALIAS,
         ha.HA_ALIAS])

    def cleanup_after_test(self):
        """Reset all class variables to their default values.
        This is needed to avoid tests to pollute subsequent tests.
        """
        TestSchedulingHACapableL3RouterServicePlugin._router_schedulers = {}
        TestSchedulingHACapableL3RouterServicePlugin._router_drivers = {}
        (TestSchedulingHACapableL3RouterServicePlugin.
         _namespace_router_type_id) = None
        (TestSchedulingHACapableL3RouterServicePlugin.
         _backlogged_routers) = set()
        (TestSchedulingHACapableL3RouterServicePlugin.
         _refresh_router_backlog) = True


class L3RouterHostingDeviceHARandomSchedulerTestCase(
        L3RoutertypeAwareHostingDeviceSchedulerTestCaseBase):

    _is_ha_tests = True

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None):
        if l3_plugin is None:
            l3_plugin = HA_L3_PLUGIN_KLASS
        if ext_mgr is None:
            ext_mgr = TestHASchedulingL3RouterApplianceExtensionManager()
        cfg.CONF.set_override('default_ha_redundancy_level', 2, group='ha')
        super(L3RouterHostingDeviceHARandomSchedulerTestCase, self).setUp(
            l3_plugin=l3_plugin, ext_mgr=ext_mgr)

    def test_ha_routers_hosted_on_different_hosting_devices(self):
        with self.subnet() as s:
            self._set_net_external(s['subnet']['network_id'])
            with self.router(external_gateway_info={
                    'network_id': s['subnet']['network_id']}) as router:
                r = router['router']
                self.l3_plugin._process_backlogged_routers()
                r_after = self._show('routers', r['id'])['router']
                self.assertIsNotNone(
                    r_after[routerhostingdevice.HOSTING_DEVICE_ATTR])
                hd_ids = {r_after[routerhostingdevice.HOSTING_DEVICE_ATTR]}
                r_rs_after = [self._show('routers', rr['id'])['router']
                              for rr in r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]]
                for rr in r_rs_after:
                    hd_id = rr[routerhostingdevice.HOSTING_DEVICE_ATTR]
                    self.assertIsNotNone(hd_id)
                    self.assertNotIn(hd_id, hd_ids)
                    hd_ids.add(hd_id)


class L3RouterHostingDeviceBaseSchedulerTestCase(
        L3RoutertypeAwareHostingDeviceSchedulerTestCaseBase):

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None):

        super(L3RouterHostingDeviceBaseSchedulerTestCase, self).setUp(
            use_ini_files=False)

    def _update_hosting_device_statuses(self, hosting_devices, statuses):
        adm_ctxt = bc.context.get_admin_context()
        core_plugin = bc.get_plugin()
        for (hosting_device, status) in zip(hosting_devices, statuses):
            hd = hosting_device['hosting_device']
            core_plugin.update_hosting_device(
                adm_ctxt, hd['id'], {'hosting_device': {'status': status}})

    def _test_get_candidates(self, templates_id, slot_capacity=5,
                             share_hosting_device=True,
                             slot_need=2, expected_candidates=None):
        r_hd_b_db = mock.MagicMock()
        r_hd_b_db.__getitem__ = lambda obj, name: fake_attrs[name]
        r_hd_b_db.share_hosting_device = share_hosting_device
        r_hd_b_db.router = mock.MagicMock()
        fake_attrs = {'template_id': templates_id,
                      'tenant_id': 'some_tenant'}
        r_hd_b_db.router.__getitem__ = lambda obj, name: fake_attrs[name]
        r_hd_b_db.router_type = mock.MagicMock()
        r_hd_b_db.router_type.slot_need = slot_need
        r_hd_b_db.router_type.__getitem__ = lambda obj, name: fake_attrs[name]
        r_hd_b_db.router_type.template = mock.MagicMock()
        r_hd_b_db.router_type.template.slot_capacity = slot_capacity
        adm_ctx = bc.context.get_admin_context()
        sched_obj = scheduler.L3RouterHostingDeviceLongestRunningScheduler()
        candidates = sched_obj.get_candidates(None, adm_ctx, r_hd_b_db)
        expected_candidates = expected_candidates or []
        self.assertEqual(len(expected_candidates), len(candidates))
        # candidates must be in correct order
        for i in range(len(candidates)):
            self.assertEqual(expected_candidates[i], candidates[i][0])

    def test_get_candidates_excludes_non_active(self):
        with self.hosting_device_template(
                host_category=HARDWARE_CATEGORY) as hdt:
            template_id = hdt['hosting_device_template']['id']
            credentials = device_manager_test_support._uuid()
            with self.hosting_device(template_id=template_id,
                                     credentials_id=credentials) as hd1,\
                    self.hosting_device(template_id=template_id,
                                        credentials_id=credentials) as hd2,\
                    self.hosting_device(template_id=template_id,
                                        credentials_id=credentials) as hd3,\
                    self.hosting_device(template_id=template_id,
                                        credentials_id=credentials) as hd4,\
                    self.hosting_device(template_id=template_id,
                                        credentials_id=credentials) as hd5,\
                    self.hosting_device(template_id=template_id,
                                        credentials_id=credentials) as hd6:
                self._update_hosting_device_statuses(
                    [hd2, hd4, hd5],
                    [c_const.HD_DEAD, c_const.HD_ERROR,
                     c_const.HD_NOT_RESPONDING])
                expected = [hd1['hosting_device']['id'],
                            hd3['hosting_device']['id'],
                            hd6['hosting_device']['id']]
                self._test_get_candidates(template_id,
                                          expected_candidates=expected)

    def test_get_candidates_excludes_admin_down(self):
        with self.hosting_device_template(
                host_category=HARDWARE_CATEGORY) as hdt:
            template_id = hdt['hosting_device_template']['id']
            credentials = device_manager_test_support._uuid()
            with self.hosting_device(template_id=template_id,
                                     credentials_id=credentials) as hd1,\
                    self.hosting_device(template_id=template_id,
                                        credentials_id=credentials,
                                        admin_state_up=False),\
                    self.hosting_device(template_id=template_id,
                                        credentials_id=credentials,
                                        admin_state_up=False),\
                    self.hosting_device(template_id=template_id,
                                        credentials_id=credentials) as hd4:
                expected = [hd1['hosting_device']['id'],
                            hd4['hosting_device']['id']]
                self._test_get_candidates(template_id,
                                          expected_candidates=expected)



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Cisco device management

Revision ID: 2921fe565328
Revises: 3c7f8bae9239
Create Date: 2015-09-28 09:00:02.923237

"""

# revision identifiers, used by Alembic.
revision = '2921fe565328'
down_revision = '3c7f8bae9239'

from alembic import op
import sqlalchemy as sa

from neutron.db import migration


def upgrade():
    op.create_table('cisco_hosting_device_templates',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('enabled', sa.Boolean(), nullable=False,
                  server_default=sa.sql.false()),
        sa.Column('host_category', sa.String(length=255), nullable=False),
        sa.Column('service_types', sa.String(length=255), nullable=True),
        sa.Column('image', sa.String(length=255), nullable=True),
        sa.Column('flavor', sa.String(length=255), nullable=True),
        sa.Column('default_credentials_id', sa.String(length=36),
                  nullable=True),
        sa.Column('configuration_mechanism', sa.String(length=255),
                  nullable=True),
        sa.Column('protocol_port', sa.Integer(), nullable=True),
        sa.Column('booting_time', sa.Integer(), nullable=True,
                  server_default='0'),
        sa.Column('slot_capacity', sa.Integer(), autoincrement=False,
                  nullable=False),
        sa.Column('desired_slots_free', sa.Integer(), autoincrement=False,
                  nullable=False, server_default='0'),
        sa.Column('tenant_bound', sa.String(length=512), nullable=True),
        sa.Column('device_driver', sa.String(length=255), nullable=False),
        sa.Column('plugging_driver', sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table('cisco_slot_allocations',
        sa.Column('template_id', sa.String(length=36), nullable=False),
        sa.Column('hosting_device_id', sa.String(length=36), nullable=False),
        sa.Column('logical_resource_type', sa.String(155), nullable=False),
        sa.Column('logical_resource_service', sa.String(155), nullable=False),
        sa.Column('logical_resource_id', sa.String(length=36), nullable=False),
        sa.Column('logical_resource_owner', sa.String(length=36),
                  nullable=False),
        sa.Column('num_allocated', sa.Integer(), autoincrement=False,
                  nullable=False),
        sa.Column('tenant_bound', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['template_id'],
                                ['cisco_hosting_device_templates.id']),
        sa.ForeignKeyConstraint(['hosting_device_id'],
                                ['cisco_hosting_devices.id']),
        sa.PrimaryKeyConstraint('logical_resource_id')
    )
    if migration.schema_has_table('cisco_hosting_devices'):
        op.add_column('cisco_hosting_devices',
                      sa.Column('template_id', sa.String(length=36),
                                nullable=False))
        op.create_foreign_key('cisco_hosting_devices_ibfk_3',
                              source_table='cisco_hosting_devices',
                              referent_table='cisco_hosting_device_templates',
                              local_cols=['template_id'], remote_cols=['id'])
        op.create_index('template_id', 'cisco_hosting_devices',
                        ['template_id'])
        op.add_column('cisco_hosting_devices',
                      sa.Column('credentials_id', sa.String(length=36),
                                nullable=True))
        op.add_column('cisco_hosting_devices',
                      sa.Column('name', sa.String(255), nullable=True))
        op.add_column('cisco_hosting_devices',
                      sa.Column('description', sa.String(255), nullable=True))
        op.add_column('cisco_hosting_devices',
                      sa.Column('management_ip_address', sa.String(255),
                                nullable=True))
        op.add_column('cisco_hosting_devices',
                      sa.Column('tenant_bound', sa.String(length=36),
                                nullable=True))
        op.add_column('cisco_hosting_devices',
                      sa.Column('auto_delete', sa.Boolean(), nullable=False,
                                server_default=sa.sql.false()))
    op.create_index(op.f('ix_cisco_hosting_device_templates_tenant_id'),
                    'cisco_hosting_device_templates', ['tenant_id'],
                    unique=False)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from oslo_serialization import jsonutils

from networking_cisco.plugins.cisco.cfg_agent.device_drivers import (
    devicedriver_api)

LOG = logging.getLogger(__name__)


class DummyRoutingDriver(devicedriver_api.RoutingDriverBase):
    """Dummy Routing Driver.

    This class emulates a routing driver without a real backing device.
    """

    def __init__(self, **device_params):
        my_device_params = device_params
        # Datetime values causes json decoding errors. So removing it locally
        if my_device_params.get('created_at'):
            del my_device_params['created_at']
        LOG.debug(jsonutils.dumps(my_device_params, sort_keys=True, indent=4))

    ###### Public Functions ########
    def router_added(self, ri):
        LOG.debug("DummyDriver router_added() called.")

    def router_removed(self, ri):
        LOG.debug("DummyDriver router_removed() called.")

    def internal_network_added(self, ri, port):
        LOG.debug("DummyDriver internal_network_added() called.")
        LOG.debug("Int port data: " + jsonutils.dumps(port, sort_keys=True,
                  indent=4))

    def internal_network_removed(self, ri, port):
        LOG.debug("DummyDriver internal_network_removed() called.")

    def external_gateway_added(self, ri, ex_gw_port):
        LOG.debug("DummyDriver external_gateway_added() called.")
        LOG.debug("Ext port data: " + jsonutils.dumps(ex_gw_port,
                                                      sort_keys=True,
                                                      indent=4))

    def external_gateway_removed(self, ri, ex_gw_port):
        LOG.debug("DummyDriver external_gateway_removed() called.")

    def enable_internal_network_NAT(self, ri, port, ex_gw_port):
        LOG.debug("DummyDriver external_gateway_added() called.")

    def disable_internal_network_NAT(self, ri, port, ex_gw_port):
        LOG.debug("DummyDriver disable_internal_network_NAT() called.")

    def floating_ip_added(self, ri, ex_gw_port, floating_ip, fixed_ip):
        LOG.debug("DummyDriver floating_ip_added() called.")

    def floating_ip_removed(self, ri, ex_gw_port, floating_ip, fixed_ip):
        LOG.debug("DummyDriver floating_ip_removed() called.")

    def routes_updated(self, ri, action, route):
        LOG.debug("DummyDriver routes_updated() called.")

    def clear_connection(self):
        LOG.debug("DummyDriver clear_connection() called.")



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from sqlalchemy import sql

from neutron.scheduler import l3_agent_scheduler

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.db.l3 import l3_models
from networking_cisco.plugins.cisco.extensions import routertype

LOG = logging.getLogger(__name__)

NEUTRON_VERSION = bc.NEUTRON_VERSION
NEUTRON_NEWTON_VERSION = bc.NEUTRON_NEWTON_VERSION
AGENT_TYPE_L3 = bc.constants.AGENT_TYPE_L3


class L3RouterTypeAwareScheduler(l3_agent_scheduler.L3Scheduler):
    """A router type aware l3 agent scheduler for Cisco router service plugin.

    It schedules Neutron routers with router type representing network
    namespace based routers to l3 agents.
    """

    def _get_unscheduled_routers(self, plugin, context):
        """Get routers with no agent binding."""
        if NEUTRON_VERSION.version[0] <= NEUTRON_NEWTON_VERSION.version[0]:
            context, plugin = plugin, context
        # TODO(gongysh) consider the disabled agent's router
        no_agent_binding = ~sql.exists().where(
            bc.Router.id == bc.rb_model.RouterL3AgentBinding.router_id)
        # Modified to only include routers of network namespace type
        ns_routertype_id = plugin.get_namespace_router_type_id(context)
        query = context.session.query(bc.Router.id)
        query = query.join(l3_models.RouterHostingDeviceBinding)
        query = query.filter(
            l3_models.RouterHostingDeviceBinding.router_type_id ==
            ns_routertype_id, no_agent_binding)
        unscheduled_router_ids = [router_id_[0] for router_id_ in query]
        if unscheduled_router_ids:
            return plugin.get_routers(
                context, filters={'id': unscheduled_router_ids})
        return []

    def _filter_unscheduled_routers(self, plugin, context, routers):
        """Filter from list of routers the ones that are not scheduled.

           Only for release < pike.
        """
        if NEUTRON_VERSION.version[0] <= NEUTRON_NEWTON_VERSION.version[0]:
            context, plugin = plugin, context
        unscheduled_routers = []
        for router in routers:
            if (router[routertype.TYPE_ATTR] !=
                    plugin.get_namespace_router_type_id(context)):
                # ignore non-namespace routers
                continue
            l3_agents = plugin.get_l3_agents_hosting_routers(
                context, [router['id']])
            if l3_agents:
                LOG.debug('Router %(router_id)s has already been '
                          'hosted by L3 agent %(agent_id)s',
                          {'router_id': router['id'],
                           'agent_id': l3_agents[0]['id']})
            else:
                unscheduled_routers.append(router)
        return unscheduled_routers

    def _get_underscheduled_routers(self, plugin, context):
        """For release >= pike."""
        underscheduled_routers = []
        max_agents_for_ha = plugin.get_number_of_agents_for_scheduling(context)

        for router, count in plugin.get_routers_l3_agents_count(context):
            if (router[routertype.TYPE_ATTR] !=
                    plugin.get_namespace_router_type_id(context)):
                # ignore non-namespace routers
                continue
            if (count < 1 or
                router.get('ha', False) and count < max_agents_for_ha):
                # Either the router was un-scheduled (scheduled to 0 agents),
                # or it's an HA router and it was under-scheduled (scheduled to
                # less than max_agents_for_ha). Either way, it should be added
                # to the list of routers we want to handle.
                underscheduled_routers.append(router)
        return underscheduled_routers

    def schedule(self, plugin, context, router, candidates=None,
                 hints=None):
        # Only network namespace based routers should be scheduled here
        ns_routertype_id = plugin.get_namespace_router_type_id(context)
        # Not very happy about these checks but since we want to work with
        # existing l3 agent scheduler they cannot be avoided
        if isinstance(router, dict):
            router_type_id = router[routertype.TYPE_ATTR]
            router_id = router['id']
        else:
            router_id = router
            r = plugin.get_router(context, router_id)
            router_type_id = r[routertype.TYPE_ATTR]
        if router_type_id == ns_routertype_id:
            # Do the traditional Neutron router scheduling
            return plugin.l3agent_scheduler.schedule(plugin, context,
                                                     router_id, candidates)
        else:
            return

    def _choose_router_agent(self, plugin, context, candidates):
        return plugin.l3agent_scheduler._choose_router_agent(plugin, context,
                                                             candidates)

    def _choose_router_agents_for_ha(self, plugin, context, candidates):
        return plugin.l3agent_scheduler._choose_router_agents_for_ha(
            plugin, context, candidates)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from neutron.common import test_lib
from neutron.tests.unit.extensions import test_l3

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import exnet_const as ext_net
from networking_cisco.plugins.cisco.device_manager.plugging_drivers.\
    hw_vlan_trunking_driver import HwVLANTrunkingPlugDriver
from networking_cisco.tests.unit.cisco.l3 import (
    test_l3_router_appliance_plugin)


class TestHwVLANTrunkingPlugDriver(
    test_l3_router_appliance_plugin.L3RouterApplianceTestCaseBase,
    test_l3.L3NatTestCaseMixin):

    # we use router types defined in .ini file.
    configure_routertypes = False
    router_type = 'ASR1k_Neutron_router'

    def setUp(self):
        super(TestHwVLANTrunkingPlugDriver, self).setUp(create_mgmt_nw=False)
        # save possible test_lib.test_config 'config_files' dict entry so we
        # can restore it after tests since we will change its value
        self._old_config_files = copy.copy(test_lib.test_config.get(
            'config_files'))
        # include config files for device manager service plugin and router
        # service plugin since we define a number of hosting device templates,
        # hosting devices and routertypes there
        self._add_device_manager_plugin_ini_file()
        self._add_router_plugin_ini_file()
        #TODO(bobmel): Fix bug in test_extensions.py and we can remove the
        # below call to setup_config()
        self.setup_config()

    def tearDown(self):
        if self._old_config_files is None:
            test_lib.test_config.pop('config_files', None)
        else:
            test_lib.test_config['config_files'] = self._old_config_files
        super(TestHwVLANTrunkingPlugDriver, self).tearDown()

    def test_create_hosting_device_resources(self):
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        plugging_driver = HwVLANTrunkingPlugDriver()
        mgmt_context = {'mgmt_nw_id': None}
        res = plugging_driver.create_hosting_device_resources(
            ctx, "some_id", tenant_id, mgmt_context, 2)
        self.assertIsNone(res['mgmt_port'])
        self.assertEqual(len(res), 1)

    def test_create_hosting_device_resources_no_mgmt_context(self):
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        plugging_driver = HwVLANTrunkingPlugDriver()
        res = plugging_driver.create_hosting_device_resources(
            ctx, "some_id", tenant_id, None, 2)
        self.assertIsNone(res['mgmt_port'], res)
        self.assertEqual(len(res), 1)

    def test_get_hosting_device_resources_by_complementary_id(self):
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        plugging_driver = HwVLANTrunkingPlugDriver()
        mgmt_context = {'mgmt_nw_id': None}
        res = plugging_driver.create_hosting_device_resources(
            ctx, "some_id", tenant_id, mgmt_context, 1)
        # ports that should not be returned
        with self.port(), self.port(device_id='uuid2'), self.port(
                tenant_id=tenant_id):
            res_get = plugging_driver.get_hosting_device_resources(
                ctx, '', 'some_id', tenant_id, None)
            self.assertIsNone(res_get['mgmt_port'])
            self.assertEqual(len(res), 1)

    def test_get_hosting_device_resources_by_device_id(self):
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        plugging_driver = HwVLANTrunkingPlugDriver()
        mgmt_context = {'mgmt_nw_id': None}
        res = plugging_driver.create_hosting_device_resources(
            ctx, "some_id", tenant_id, mgmt_context, 1)
        # update attributes of created ports to fake what Nova updates
        hd_uuid = 'hd_uuid1'
        update_spec = {'port': {'device_id': hd_uuid,
                                'device_owner': 'nova'}}
        for hd_port in self._list('ports')['ports']:
            self._update('ports', hd_port['id'], update_spec)
        # ports that should not be returned
        with self.port(), self.port(device_id='uuid2'), self.port(
                tenant_id=tenant_id), self.port(tenant_id=tenant_id,
                                                device_owner='other_uuid'):
            res_get = plugging_driver.get_hosting_device_resources(
                ctx, hd_uuid, 'some_id', tenant_id, None)
            self.assertIsNone(res_get['mgmt_port'])
            self.assertEqual(len(res), 1)

    def test_delete_hosting_device_resources(self):
        tenant_id = 'some_tenant_id'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        mgmt_context = {'mgmt_nw_id': None}
        plugging_driver = HwVLANTrunkingPlugDriver()
        res = plugging_driver.create_hosting_device_resources(
            ctx, "some_id", tenant_id, mgmt_context, 2)
        nets = self._list('networks')
        self.assertEqual(len(nets['networks']), 0)
        subnets = self._list('subnets')
        self.assertEqual(len(subnets['subnets']), 0)
        ports = self._list('ports')
        self.assertEqual(len(ports['ports']), 0)
        # avoid passing the mgmt port twice in argument list
        mgmt_port = res['mgmt_port']
        del res['mgmt_port']
        plugging_driver.delete_hosting_device_resources(
            ctx, tenant_id, mgmt_port, **res)
        nets = self._list('networks')['networks']
        # no networks and subnets should remain
        self.assertEqual(len(nets), 0)
        subnets = self._list('subnets')['subnets']
        self.assertEqual(len(subnets), 0)
        ports = self._list('ports')
        self.assertEqual(len(ports['ports']), 0)

    def test_extend_hosting_port_info_adds_segmentation_id_internal(self):
        hosting_info = {}
        fake_port_db_obj = mock.MagicMock()
        fake_port_db_obj.hosting_info = mock.MagicMock()
        fake_port_db_obj.hosting_info.segmentation_id = 50
        fake_port_db_obj.device_owner = bc.constants.DEVICE_OWNER_ROUTER_INTF
        hosting_device = {'id': '00000000-0000-0000-0000-000000000002'}
        tenant_id = 'tenant_uuid1'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        plugging_driver = HwVLANTrunkingPlugDriver()
        if bc.NEUTRON_VERSION < bc.NEUTRON_PIKE_VERSION:
            fake_port_db_obj.networks.external = None
        else:
            uuid = 'some_uuid'
            fake_port_db_obj.network_id = uuid
            get_nw_mock = mock.MagicMock(
                side_effect=lambda r_ctx, nw_id: (
                    {ext_net.EXTERNAL: False} if nw_id == uuid else {}))
            mock.patch.object(plugging_driver._core_plugin, 'get_network',
                              get_nw_mock).start()
        plugging_driver.extend_hosting_port_info(ctx, fake_port_db_obj,
                                                 hosting_device, hosting_info)
        self.assertEqual(hosting_info['physical_interface'],
                         'GigabitEthernet/1/0/1')
        self.assertEqual(hosting_info['segmentation_id'], 50)

    def test_extend_hosting_port_info_adds_segmentation_id_external(self):
        hosting_info = {}
        fake_port_db_obj = mock.MagicMock()
        fake_port_db_obj.hosting_info = mock.MagicMock()
        fake_port_db_obj.hosting_info.segmentation_id = 40
        fake_port_db_obj.device_owner = bc.constants.DEVICE_OWNER_ROUTER_GW
        hosting_device = {'id': '00000000-0000-0000-0000-000000000002'}
        tenant_id = 'tenant_uuid1'
        ctx = bc.context.Context('', tenant_id, is_admin=True)
        plugging_driver = HwVLANTrunkingPlugDriver()
        if bc.NEUTRON_VERSION < bc.NEUTRON_PIKE_VERSION:
            fake_port_db_obj.networks.external = {'external': True}
        else:
            uuid = 'some_uuid'
            fake_port_db_obj.network_id = uuid
            get_nw_mock = mock.MagicMock(
                side_effect=lambda r_ctx, nw_id: (
                    {ext_net.EXTERNAL: True} if nw_id == uuid else {}))
            mock.patch.object(plugging_driver._core_plugin, 'get_network',
                              get_nw_mock).start()
        plugging_driver.extend_hosting_port_info(ctx, fake_port_db_obj,
                                                 hosting_device, hosting_info)
        plugging_driver.extend_hosting_port_info(ctx, fake_port_db_obj,
                                                 hosting_device, hosting_info)
        self.assertEqual(hosting_info['physical_interface'],
                         'GigabitEthernet/2/0/1')
        self.assertEqual(hosting_info['segmentation_id'], 40)

    def _update_provider_net_info(self, res_list, fields):
        for res in res_list:
            pv_info = self._pv_info['vlan'].get(res['id'])
            if pv_info is None:
                pv_info = self._pv_info['vxlan'].get(res['id'])
                if pv_info is None:
                    nw_type = self._network_type
                    if not self._pv_info[nw_type]:
                        tag = {'vlan': 50, 'vxlan': 7000}[nw_type]
                        pv_info = {'nw_type': nw_type, 'tag': tag}
                        self._pv_info[nw_type][res['id']] = pv_info
            if pv_info is None:
                tag = max([i['tag']
                           for i in self._pv_info[nw_type].values()]) + 1
                pv_info = {'nw_type': nw_type, 'tag': tag}
                self._pv_info[nw_type][res['id']] = pv_info
            res[bc.provider_net.NETWORK_TYPE] = pv_info['nw_type']
            res[bc.provider_net.SEGMENTATION_ID] = pv_info['tag']
            if fields is not None:
                for attr in list(res.keys()):
                    if attr not in fields:
                        del res[attr]

    def _mocked_get_network(self, context, id, fields=None):
        res = self.real_get_network(context, id)
        self._update_provider_net_info([res], fields)
        return res

    def _mocked_get_networks(self, *args, **kwargs):
        if len(args) >= 3:
            fields = args[2]
            list_args = [i for i in args]
            list_args[2] = None
            args = list_args
        else:
            fields = kwargs.pop('fields', None)
        res_list = self.real_get_networks(*args, **kwargs)
        self._update_provider_net_info(res_list, fields)
        return res_list

    def _test_allocate_hosting_port(self, test_info1):

        def _validate_allocation(plugin, ctx, r, port_db, test_info,
                                 i, hd, plugging_driver):
            binding_db = plugin._allocate_hosting_port(
                ctx, r['id'], port_db, hd['id'], plugging_driver)
            self.assertIsNotNone(binding_db.hosting_port_id)
            self.assertEqual(binding_db.hosting_port_id,
                             port_db.id)
            self.assertEqual(binding_db.segmentation_id,
                             test_info['vlan_tags'][i])

        plugging_driver = HwVLANTrunkingPlugDriver()
        with self.subnet() as subnet1:
            sn1 = subnet1['subnet']
            ext_net_id = sn1['network_id']
            self._set_net_external(ext_net_id)
            gw_info = {'network_id': ext_net_id}
            with self.router(external_gateway_info=gw_info,
                             tenant_id=sn1['tenant_id']) as router1:
                r1 = router1['router']
                hds = self._list('hosting_devices')['hosting_devices']
                hd = hds[0]
                self._pv_info = {'vlan': {}, 'vxlan': {}}
                self._network_type = test_info1['network_types'][0]
                self.real_get_network = self.core_plugin.get_network
                self.real_get_networks = self.core_plugin.get_networks
                with mock.patch.object(self.core_plugin, 'get_network') as m1,\
                        mock.patch.object(self.core_plugin,
                                          'get_networks') as m2:
                    m1.side_effect = self._mocked_get_network
                    m2.side_effect = self._mocked_get_networks
                    u1_ctx = bc.context.Context('', r1['tenant_id'],
                                             is_admin=True)
                    gw_port_db = self.core_plugin._get_ports_query(
                        u1_ctx, filters={'network_id': [ext_net_id]}).one()
                    _validate_allocation(
                        self.l3_plugin, u1_ctx, r1, gw_port_db,
                        test_info1, 0, hd, plugging_driver)
                    for i in range(1, len(test_info1['network_types'])):
                        cidr = '1.0.' + str(i) + '.0/24'
                        with self.subnet(cidr=cidr) as subnet2:
                            sn2 = subnet2['subnet']
                            itfc_info = self._router_interface_action(
                                'add', r1['id'], sn2['id'], None)
                            self._network_type = test_info1['network_types'][i]
                            port_db = self.core_plugin._get_port(
                                u1_ctx, itfc_info['port_id'])
                            _validate_allocation(
                                self.l3_plugin, u1_ctx, r1, port_db,
                                test_info1, i, hd, plugging_driver)

    def test_allocate_hosting_port_vlan_network_all_unused(self):
        self._test_allocate_hosting_port({'network_types': ['vlan'],
                                          'vlan_tags': [50]})

    def test_allocate_hosting_port_vlan_network_vlan_already_allocated(self):
        self._test_allocate_hosting_port({'network_types': ['vlan', 'vlan'],
                                          'vlan_tags': [50, 51]})

    def test_allocate_hosting_port_vlan_network_not_found_failure(self):
        with self.subnet() as subnet1:
            sn1 = subnet1['subnet']
            ext_net_id = sn1['network_id']
            self._set_net_external(ext_net_id)
            gw_info = {'network_id': ext_net_id}
            with self.router(external_gateway_info=gw_info,
                             tenant_id=sn1['tenant_id']) as router1:
                r1 = router1['router']
                plugging_driver = HwVLANTrunkingPlugDriver()
                u_ctx = bc.context.Context('', r1['tenant_id'], is_admin=True)
                gw_port_db = self.core_plugin._get_ports_query(
                        u_ctx, filters={'network_id': [ext_net_id]}).one()
                allocations = plugging_driver.allocate_hosting_port(
                    u_ctx, r1['id'], gw_port_db, 'vlan', 'non_existant_uuid')
                self.assertIsNone(allocations)



# Copyright 2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class BaseDriver(object):

    """Base Driver class for FW driver classes. """

    # def __init__(self):
    #    Pass

    @abc.abstractmethod
    def initialize(self):
        """Initialize method. """
        pass

    @abc.abstractmethod
    def populate_event_que(self):
        """Populate Event Queue. """
        pass

    @abc.abstractmethod
    def populate_dcnm_obj(self):
        """Populate DCNM Obj. """
        pass

    @abc.abstractmethod
    def network_create_notif(self, tenant_id, tenant_name, cidr):
        """Network Create Notification. """
        pass

    @abc.abstractmethod
    def network_delete_notif(self, tenant_id, tenant_name, net_id):
        """Network Delete Notification. """
        pass

    @abc.abstractmethod
    def is_device_virtual(self):
        """Return False if device is physical, True otherwise. """
        pass

    @abc.abstractmethod
    def get_name(self):
        """Return the name of the driver service. """
        pass

    @abc.abstractmethod
    def get_max_quota(self):
        """Retrieves the maximum number of FW that could be created. """
        pass

    @abc.abstractmethod
    def create_fw(self, tenant_id, data):
        """Create the Firewall. """
        pass

    @abc.abstractmethod
    def delete_fw(self, tenant_id, data):
        """Delete the Firewall. """
        pass

    @abc.abstractmethod
    def modify_fw(self, tenant_id, data):
        """Modify the Firewall. """
        pass
